#!/usr/bin/env python

from __future__ import print_function, absolute_import

# -*- coding: utf-8 -*-
__author__ = "perfguru87@gmail.com"
__copyright__ = "Copyright 2018, The PerfTracker project"
__license__ = "MIT"

"""
A set of various free function helpers
"""

import os
import re
import sys
import logging

try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse

if sys.version_info[0] < 3:
    import httplib
else:
    import http.client as httplib


reRawCookie = re.compile(r"[Ss]et-[Cc]ookie:\s*(.*)\r\n")


def gen_urls_from_index_file(path):
    urls = []
    for p in path:
        if os.path.exists(p):
            if os.path.isdir(p):
                import glob
                urls += ["file://%s" % f for f in glob.glob(os.path.join(p, "*.htm*"))]
            else:
                urls += ["file://" + p]
        else:
            from .browser_python import BrowserPython
            b = BrowserPython()
            if not p.startswith("http"):
                p = "http://%s" % p
            txt = b.http_get(p)

            if txt.find("<title>Index of") >= 0:
                r = re.compile(r">([-\w]+\.htm[l]?)<")
                for line in txt.splitlines():
                    m = r.search(line)
                    if m:
                        urls.append("%s%s" % (p, m.group(1)))
            else:
                urls += [p]
    return urls


def get_val(d, keys, default):
    v = d
    try:
        for k in keys:
            v = v[k]
    except KeyError:
        return default
    return v


_parsed_url = {}


def parse_url(url, server=False, args=False):
    global _parsed_url

    if not server and not args and url in _parsed_url:
        return _parsed_url[url]

    scheme = url.split(":")[0]

    _url = url if scheme in ("http", "https", "file") else "http://%s" % url

    p = urlparse(_url)
    scheme = p.scheme if p.scheme and p.scheme != "" else "http"

    if not p.netloc and not url.startswith("file://"):
        msg = "Can't parse network location for the following url: %s" % url
        logging.error(msg)
        from .browser_base import BrowserExc
        raise BrowserExc(msg)

    if server:
        return "%s://%s" % (scheme, p.netloc)

    path = p.path if p.path else "/"

    if args:
        pfx = "%s://%s" % (scheme, p.netloc)
        if url.startswith(pfx):
            return scheme, p.netloc, url[len(pfx):]

    netloc = p.netloc
    drop_port = ":443" if scheme == 'https' else ":80"
    if 'http' in scheme and p.netloc[-len(drop_port):] == drop_port:
        netloc = netloc.split(":")[0]

    _parsed_url[url] = scheme, netloc, path
    return _parsed_url[url]


def get_common_url_prefix(urls):
    netlocs = set()
    for url in urls:
        prot, netloc, _ = parse_url(url)
        netlocs.add("%s://%s" % (prot, netloc))

    if len(netlocs) == 1:
        return next(iter(netlocs))
    return ""


def extract_cookies(http_response):
    """
    Httplib specific function: HTTPResponse has some issue regarding
    to multiple 'set-cookie' headers. It just do ','.join(cookies), and
    there will be a trouble if cookie has ',' inside body (and it usually has!)
    So in case of multiple 'set-cookie' headers we should retrieve them
    manually (not so hard, actually)
    :param http_response: HTTPResponse object
    :return: list of strings with cookies
    """
    cookies = []
    if type(http_response) in (httplib.HTTPConnection, httplib.HTTPSConnection):
        raw_cookies = http_response.msg.getallmatchingheaders('set-cookie')
        if not raw_cookies:
            return []
        if len(raw_cookies) == 1:
            return [http_response.getheader('set-cookie')]

        for raw_cookie in raw_cookies:
            m = reRawCookie.search(raw_cookie)
            cookies.append(m.group(1))
        return cookies
    # for other implementations (e.g. pycurl just assume list of response headers is correct)
    return [value for hdr, value in http_response.getheaders() if hdr.lower() == "set-cookie"]



#!/usr/bin/env python

from __future__ import print_function, absolute_import

# -*- coding: utf-8 -*-
__author__ = "perfguru87@gmail.com"
__copyright__ = "Copyright 2018, The PerfTracker project"
__license__ = "MIT"

"""
The Chrome browser helper
"""

import logging
import re
import os
import datetime
import json
import sys
import tempfile
import atexit
import shutil

from .utils import parse_url, get_val
from .browser_base import BrowserExc
from .browser_webdriver import BrowserWebdriver, abort
from .page import PageEvent, PageRequest

try:
    from selenium import webdriver
    from selenium.common.exceptions import WebDriverException
    import psutil
except ImportError as e:
    abort(e)

reChromeEvents = re.compile(r"DEVTOOLS EVENT ([\w\.]+)([\n\W\w]*)")


class BrowserChrome(BrowserWebdriver):
    engine = "chrome"
    skip_urls = ["google.com", "gstatic.com"]

    def _skip_url(self, page, url):
        if url and url.startswith("chrome-search:"):
            return True

        return BrowserWebdriver._skip_url(self, page, url)

    def _add_event(self, page, name, params):
        if logging.getLogger().getEffectiveLevel() == logging.DEBUG:
            self.log_debug(name + '\n' + json.dumps(params, indent=4))

        if not self._ts_offset:
            # 1. some Chrome browsers log 'wallTime'
            try:
                self._ts_offset = 1000 * (float(params['wallTime']) - float(params['timestamp']))
                self.log_debug("_ts_offset = %d, detected from request wallTime timestamp" % self._ts_offset)
            except KeyError:
                pass

            if name == "Network.requestWillBeSent":
                # 2. but some do not log 'wallTime', so lets align browser timestamp with time of first request
                try:
                    _, netloc, _ = parse_url(params['documentURL'])
                    if netloc == self._first_navigation_netloc:
                        self._ts_offset = 1000 * (self._first_navigation_ts - float(params['timestamp']))
                        self.log_debug("_ts_offset = %d, detected from requestWillBeSent event" % self._ts_offset)
                except KeyError:
                    raise BrowserExc("can't determine timestamp offset between wall time and browser log timestamps")

            if not self._ts_offset:
                self.log_debug("skip event '%s'" % name)
                return

        try:
            ts = 1000 * float(params['timestamp']) + self._ts_offset
            page.process_activity(name, ts)
        except KeyError:
            ts = None

        e = PageEvent(name, params)
        if name == "Network.requestWillBeSent":
            if params['request']['url'].startswith("data:"):
                return

            req_id = params['requestId']
            if page.get_request(req_id):
                self.log_debug("WARNING: Response has been already received for req %s" % req_id)

            url = params['request']['url']
            if self._skip_url(page, url):
                self.log_debug("ignore/skip request to: %s" % url)
                return

            r = PageRequest(page, req_id)
            r.method = params['request']['method']
            r.header = params['request']['headers']
            if r.method == "POST" and 'postData' in params['request']:
                r.params = params['request']['postData']
            r.url = url
            r.start(ts)
            page.add_request(r)

        elif name in ("Network.responseReceived", "Network.dataReceived", "Network.requestServedFromCache",
                      "Network.loadingFinished", "Network.loadingFailed"):
            req_id = params['requestId']
            if not page.get_request(req_id):
                url = get_val(params, ['response', 'url'], '')
                if url.startswith("data:"):
                    return

                if self._skip_url(page, url):
                    self.log_debug("ignore/skip response from: %s" % url)
                    return

                self.log_debug("WARNING: got '%s' for unsent request %s" % (name, req_id))
                return

            r = page.get_request(req_id)

            if get_val(params, ['response', 'fromDiskCache'], False):
                r.cached = True

            if name == "Network.responseReceived":
                r.set_type(params['type'])
                r.connection_reused = get_val(params, ['response', 'connectionReused'], False)
                r.status = get_val(params, ['response', 'status'], 'unknown')
                r.keepalive = get_val(params, ['response', 'headers', 'Connection'], '') in ("Keep-Alive", "keep-alive")
                r.gzipped = get_val(params, ['response', 'headers', 'Content-Encoding'], '') == "gzip"
                r.content_length = int(get_val(params, ['response', 'headers', 'Content-Length'], 0))

                # response header normally contains cookie
                cookie = get_val(params, ['response', 'requestHeaders', 'Cookie'], None)
                if cookie:
                    r.header['Cookie'] = cookie

                if not r.status:
                    self.log_debug("WARNING: Can't parse status for request %s" % req_id)
                r.complete(ts)

            elif name == "Network.dataReceived":
                r.length += int(params['dataLength'])
                r.completed = False

            elif name == "Network.requestServedFromCache":
                r.cached = True
                r.complete(ts)

            elif name == "Network.loadingFinished":
                r.complete(ts)

            elif name == "Network.loadingFailed":
                r.status = params['errorText']
                r.set_type(params['type'])
                r.complete(ts)

    def _browser_parse_logs(self, page, logs):
        if logging.getLogger().getEffectiveLevel() == logging.DEBUG:
            self.log_debug("LOGS: %s" % json.dumps(logs, indent=4))

        if 'value' not in logs:
            return
        for val in logs['value']:
            if val['message'].startswith("{"):
                try:
                    j = json.loads(val['message'])
                    self._add_event(page, j['message']['method'], j['message']['params'])
                    continue
                except ValueError:
                    pass

            if type(val['message']) == str:
                v = val['message']
            else:
                v = val['message'].decode('string-escape')
            if v.startswith("DEVTOOLS EVENT"):
                m = reEvents.match(v)
                if m:
                    self._add_event(page, m.group(1), m.group(2))

    def _browser_get_events(self, page):
        self._browser_parse_logs(page, self.driver.execute('getLog', {'type': 'browser'}))
        self._browser_parse_logs(page, self.driver.execute('getLog', {'type': 'performance'}))

    def _browser_get_rss_kb(self):
        # w/a for docker
        if self.remote_connstring:
            return 0
        rss = 0
        proc = psutil.Process(self.pid)
        if not proc:
            self.log_error("Something went wrong and browser is dead!")
            sys.exit(-1)
        children = proc.get_children() if hasattr(proc, 'get_children') else proc.children()
        for p in children:
            rss += p.memory_info().rss
        return rss / 1024

    def browser_start(self):
        d = tempfile.mkdtemp()
        self.log_debug("Temporary user data directory: %s" % d)
        atexit.register(shutil.rmtree, d)

        options = webdriver.ChromeOptions()
        options.add_argument("--window-size=%d,%d" % (self.resolution[0], self.resolution[1]))
        options.add_argument("--no-sandbox")
        options.add_argument("--user-data-dir=%s" % d)
        options.add_argument("--disable-setuid-sandbox")

        descaps = options.to_capabilities()

        # chromedriver capability was renamed to goog:loggingPrefs, as required by W3C standard
        descaps['goog:loggingPrefs'] = {'performance': 'DEBUG', 'browser': 'ALL', 'driver': 'ALL'}

        # w/a for hanging Chrome, see https://github.com/SeleniumHQ/docker-selenium/issues/87
        os.environ["DBUS_SESSION_BUS_ADDRESS"] = "/dev/null"

        if self.remote_connstring:
            try:
                self.driver = webdriver.Remote(self.remote_connstring,
                                               desired_capabilities=descaps)
                return 1
            except WebDriverException as e:
                abort(e)
        else:
            try:
                self.driver = webdriver.Chrome(desired_capabilities=descaps,
                                               service_args=['--verbose', '--log-path=%s' % self.log_path])
                return self.driver.service.process.pid
            except WebDriverException as e:
                abort(e)

    def navigation_reset(self):
        self.navigate_to("chrome://version/")


##############################################################################
# Autotests
##############################################################################


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    try:
        b = BrowserChrome()
    except BrowserExc as e:
        sys.exit(-1)
    b.navigate_to("https://example.com/")
    b.browser_stop()



#!/usr/bin/env python

from __future__ import print_function, absolute_import

# -*- coding: utf-8 -*-
__author__ = "perfguru87@gmail.com"
__copyright__ = "Copyright 2018, The PerfTracker project"
__license__ = "MIT"

"""
Skeleton for webdriver based browser
"""

import logging
import re
import os
import datetime
import time
import json
import sys
import tempfile
import atexit
import shutil

from perftrackerlib.helpers import timeparser
from perftrackerlib.helpers import largelogfile

from .utils import parse_url, get_val
from .browser_base import BrowserBase, BrowserExc, BrowserExcTimeout, BrowserExcNotImplemented
from .page import Page, PageEvent, PageRequest, PageTimeline


if sys.version_info[0] < 3:
    from urllib2 import URLError
else:
    from urllib.error import URLError


def help():
    ret = ["-" * 80,
           "The browser library requires:",
           "* Install python > 2.6 and *selenium*, *pyvirtualdisplay* and *psutil* modules",
           "     # sudo pip install selenium pyvirtualdisplay psutil"
           ]

    if sys.platform == "linux2":
        ret += ["     # yum install Xvfb"]

    ret += ["* To enable Chrome:",
            "     install Chrome browser version >= 40:",
            "        yum install https://dl.google.com/linux/direct/google-chrome-stable_current_x86_64.rpm",
            "     install chromedriver from here http://chromedriver.storage.googleapis.com/index.html",
            "     copy the crhromedriver to /usr/bin/ and make it executable",
            "* To enable Firefox:",
            "     install Firefox browser version >= 31",
            ]
    return ret


def abort(msg):
    print("Error: " + str(msg), file=sys.stderr)
    print("\n".join(help()), file=sys.stderr)
    sys.exit(-1)


try:
    from selenium import webdriver
    from selenium.common.exceptions import WebDriverException
    from selenium.common.exceptions import NoSuchElementException
    from selenium.common.exceptions import ElementNotVisibleException
    from selenium.common.exceptions import StaleElementReferenceException
    import psutil
except ImportError as e:
    abort(e)


class BrowserWebdriver(BrowserBase):
    skip_urls = []

    def __init__(self, *args, **kwargs):
        BrowserBase.__init__(self, *args, **kwargs)
        self._first_navigation_ts = None
        self._first_navigation_netloc = None
        self._ts_offset = None

    def _skip_url(self, page, url):
        if not url:
            return False

        _, req_netloc, _ = parse_url(url)

        for su in self.skip_urls:
            if su in req_netloc:
                _, page_netloc, _ = parse_url(page.url)
                if not any(x in page_netloc for x in self.skip_urls):
                    self.log_debug("skipping URL %s" % req_netloc)
                    return True
        return False

    def _browser_clear_caches(self):
        BrowserBase._browser_clear_caches(self)
        self.driver.quit()
        self.pid = self.browser_start()

    def _browser_navigate(self, location, cached=True, name=None):
        url = location.url if isinstance(location, Page) else location
        real_navigation = self._http_get(url)
        return Page(self, url, cached, name=name, real_navigation=real_navigation)

    def _browser_wait(self, page, timeout=None):

        self.log_info("_browser_wait()...")

        if timeout is None:
            timeout = self.nav_timeout

        start = time.time()
        while time.time() - start < timeout / 2:
            time.sleep(0.2)
            if self.driver.execute_script("return window.performance.timing.loadEventEnd"):
                break
            # onload event has not been processed yet, so need to wait and retry
            self.log_info("Waiting for loadEventEnd ... ")

        while time.time() - start < timeout:
            time.sleep(self.ajax_threshold)

            # hack. Execute something in browser context to flush logs...
            self.driver.execute_script("return window.performance.timing.loadEventEnd")

            self._browser_get_events(page)

            ir = page.get_incomplete_reqs()
            if not ir:
                break
            self.log_info("Waiting for incomplete requests:\n    %s" %
                          ("\n    ".join(["%s - %s" % (r.id, r.url) for r in ir])))

        if time.time() - start >= timeout:
            if not self.driver.execute_script("return window.performance.timing.loadEventEnd"):
                self.log_error("Page '%s' load timeout, window.performance.timing.loadEventEnd = 0" % page.url)

            ir = page.get_incomplete_reqs()
            if ir:
                self.log_error("Can't wait for page '%s' load completion, "
                               "see '%s' for details\nincomplete requests:\n    %s" %
                               (page.url, self.log_path, "\n    ".join(["%s - %s" % (r.id, r.url) for r in ir])))

        page.complete(self)

    def _browser_warmup_page(self, location, name=None):
        self.navigate_to(location, cached=False, stats=False, name=name)

    def _browser_display_init(self, headless, resolution):
        if headless:
            try:
                from pyvirtualdisplay import Display
            except ImportError as e:
                abort(e)
            self.display = Display(visible=0, size=resolution)
            self.display.start()
        else:
            self.display = None

    def _browser_execute_script(self, js):
        val = self.driver.execute_script("return %s" % js)
        self.log_debug("%s = %s" % (js, val))
        return val

    def browser_get_name(self):
        c = self.driver.capabilities
        return c['browserName']

    def browser_get_version(self):
        c = self.driver.capabilities
        return self._get_val(c, ['version', 'browserVersion'])

    def browser_get_platform(self):
        c = self.driver.capabilities
        return self._get_val(c, ['platform', 'platformName'])

    def browser_get_screenshot_as_file(self, filename):
        self.driver.get_screenshot_as_file(filename)

    def browser_get_page_timeline(self, page):

        values = {}
        for t in PageTimeline.types:
            if t in PageTimeline.jstypes:
                js = "window.performance.timing.%s" % PageTimeline.jstypes[t]
                values[t] = self._browser_execute_script(js)

        return PageTimeline(page, values)

#    def browser_set_session(self, domain, session_id):
#        self._http_get(domain)
#        self.driver.add_cookie({'name': 'sessionid', 'value': session_id})

    def browser_get_current_url(self):
        return self.driver.current_url

    def browser_get_screenshot(self, filename):
        self.driver.get_screenshot_as_file(filename)

    def browser_stop(self):
        try:
            if self.driver:
                self.driver.quit()
                self.driver = None
            if self.display:
                self.display.stop()
                self.display = None
        except URLError:
            pass

    def _xpath_click(self, xpath):
        exc = None

        # take into account possible replacements of %23/#
        xpaths = [xpath]
        if "%23" in xpath:
            xpaths.append(xpath.replace("%23", "#"))
        if "#" in xpath:
            xpaths.append(xpath.replace("#", "%23"))

        for x in xpaths:
            self.log_debug("Looking for xpath: %s ..." % x)
            try:
                el = self.driver.find_element_by_xpath(x)
                el.click()
                self.log_debug("Looking for xpath: %s ... OK" % x)
                return
            except NoSuchElementException as e:
                self.log_debug("Looking for xpath: %s ... Failed, no such element" % x)
                exc = e
            except ElementNotVisibleException as e:
                self.log_warning("Looking for xpath: %s ... Failed, element not visible" % x)
                exc = e

        self.log_error("NoSuchElementException, xpath: %s, see debug log" % xpath)
        self.log_debug("page source:\n%s" % self.driver.page_source.encode('ascii', 'ignore'))
        raise BrowserExc(e)

    def _http_get(self, url, validator=None):
        self.log_debug("Execute GET request: %s" % url)

        if not self._first_navigation_ts:
            self._first_navigation_ts = time.time()
            _, self._first_navigation_netloc, _ = parse_url(url)

        ar = url.split("^")
        if len(ar) > 1:
            self._xpath_click(ar[1])
            return False

        try:
            self.driver.get(url)
        except WebDriverException as e:
            raise BrowserExc(e)
        return True

    @staticmethod
    def _get_val(d, keys):
        for key in keys:
            if key in d:
                return d[key]
        return "unknown"

    def print_browser_info(self):
        c = self.driver.capabilities
        self.print_stats_title("Browser summary")
        print("  - platform: %s" % self.browser_get_platform())
        print("  - browser:  %s %s" % (self.browser_get_name(), self.browser_get_version()))
        print("  - PID:      %d" % self.pid)
        print("  - log file: %s" % self.log_path)

    def print_log_file_path(self):
        self.print_stats_title("Browser log file")
        print("  %s" % self.log_path)

    # === virtual methods that must be implemented in every webdriver-based browser === #

    def _browser_parse_logs(self, page, logs):
        raise BrowserExcNotImplemented()

    def _browser_get_events(self, page):
        raise BrowserExcNotImplemented()

    # === webdriver specific === #

    def dom_wait_element_stale(self, el, timeout_s=None, name=None):
        start_time = time.time()

        if timeout_s is None:
            timeout_s = self.nav_timeout

        # http://www.obeythetestinggoat.com/how-to-get-selenium-to-wait-for-page-load-after-a-click.html
        while time.time() < start_time + timeout_s:
            try:
                el.find_elements_by_id('doesnt-matter')
                pass
            except StaleElementReferenceException:
                break
            time.sleep(0.1)

        if time.time() > start_time + timeout_s:
            msg = "DOM element '%s' click() timeout: %.1fs" % (name, time.time() - start_time)
            self.log_error(msg)
            raise BrowserExcTimeout(msg)

    def dom_click(self, el, timeout_s=None, name=None, wait_callback=None, wait_callback_obj=None):
        self.log_debug("dom_click(%s, %s)" % (str(el), str(name)))

        if timeout_s is None:
            timeout_s = self.nav_timeout

        p = Page(self, self.browser_get_current_url(), True, name=name, real_navigation=False)
        p.start()

        # 1. click on the element

        old_page = self.driver.find_element_by_tag_name('html')
        el.click()

        # 2. wait for selenium onclick completion

        if wait_callback:
            self.log_debug("wait callback: %s, %s" % (str(wait_callback.__name__), str(wait_callback_obj)))
            wait_callback(wait_callback_obj, el, timeout_s, name)
        else:
            self.log_debug("wait stale: %s, %s, %s" % (el, timeout_s, name))
            self.dom_wait_element_stale(el, timeout_s, name)

        # 3. wait for ajax completion, because browser URL can be update only after that

        self._browser_wait(p, timeout=timeout_s)
        p.url = self.browser_get_current_url()

        time.sleep(0.2)

    def dom_find_element_by_id(self, id):
        try:
            return self.driver.find_element_by_id(id)
        except NoSuchElementException as e:
            raise BrowserExc(e)

    def dom_find_element_by_name(self, name):
        try:
            return self.driver.find_element_by_name(name)
        except NoSuchElementException as e:
            raise BrowserExc(e)

    def dom_find_element_by_xpath(self, xpath):
        try:
            return self.driver.find_element_by_xpath(xpath)
        except NoSuchElementException as e:
            raise BrowserExc(e)

    def dom_find_frames(self):
        frames = []
        for name in ("frame", "iframe"):
            try:
                frames += self.driver.find_elements_by_tag_name(name)
            except NoSuchElementException as e:
                pass
        return frames

    def dom_switch_to_frame(self, frame):
        self.log_info("Switching to frame %s" % frame)
        return self.driver.switch_to.frame(frame)

    def dom_switch_to_default_content(self):
        self.log_info("Switching to default content")
        return self.driver.switch_to.default_content()

    def dom_send_keys(self, el, keys):
        val = el.get_attribute('value')
        if val != '':  # clear initial value
            self.log_info("Element value is not empty, clear content...")
            self.driver.execute_script("arguments[0].value = ''", el)
            time.sleep(2.0)

        for ch in keys:
            el.send_keys(ch)
            time.sleep(0.2)
        val = el.get_attribute('value')
        if val == keys:
            return True

        self.log_warning("Bogus selenium send_keys(). Entered: '%s', "
                         "but see: '%s', using set_attribute()..." % (keys, val))
        time.sleep(2.0)
        self.driver.execute_script("arguments[0].value = '%s'" % keys, el)
        time.sleep(2.0)
        val = el.get_attribute('value')
        if val == keys:
            self.log_info("Ok, set_attribute() works fine")
            return True

        self.log_error("Bogus selenium send_keys() and set_attribute(), can't enter value into the element")
        return False

    # === some predefined scenarios === #

    def _do_send_keys(self, title, keys, tag_names, tag_ids):
        for tag, name in tag_names:
            try:
                el = self.dom_find_element_by_name(name)
                if el.tag_name != tag:
                    continue
                if not self.dom_send_keys(el, keys):
                    self.log_error("Couldn't enter %s" % title)
                    return False
                return True
            except BrowserExc as e:
                pass

        for tag, id in tag_ids:
            try:
                el = self.dom_find_element_by_id(id)
                if el.tag_name != tag:
                    continue
                if not self.dom_send_keys(el, keys):
                    self.log_error("Couldn't enter %s" % title)
                    return False
                return True
            except BrowserExc as e:
                pass

        self.log_info("Couldn't find %s input field" % title)
        return False

    def _do_login(self, url, user, password, login_form, timeout_s=None):
        if not self._do_send_keys('user name', user, login_form.user_tags, login_form.user_ids):
            return False

        time.sleep(1)

        if not self._do_send_keys('password', password, login_form.pass_tags, login_form.pass_ids):
            return False

        time.sleep(1)

        submit_form_found = False
        for tag, name in login_form.sbmt_tags:
            try:
                el = self.dom_find_element_by_name(name)
                if el.tag_name != tag:
                    continue
                submit_form_found = True
                self.dom_click(el, name=name, timeout_s=timeout_s)

                try:
                    el = self.dom_find_element_by_name(name)
                except BrowserExc:
                    self.log_info("Login succeed")
                    return True

            except BrowserExc as e:
                pass

        for tag, id in login_form.sbmt_ids:
            try:
                el = self.dom_find_element_by_id(id)
                if el.tag_name != tag:
                    continue
                submit_form_found = True
                self.dom_click(el, name=id, timeout_s=timeout_s)

                try:
                    el = self.dom_find_element_by_id(id)
                except BrowserExc:
                    self.log_info("Login succeed")
                    return True

            except BrowserExc as e:
                pass

        for x in login_form.sbmt_xpath:
            try:
                el = self.dom_find_element_by_xpath(x)
                submit_form_found = True
                self.dom_click(el, name=id, timeout_s=timeout_s)

                try:
                    el = self.dom_find_element_by_xpath(x)
                except BrowserExc:
                    self.log_info("Login succeed")
                    return True

            except BrowserExc as e:
                pass

        if not submit_form_found:
            self.log_info("Couldn't find login submit form")

        self.log_info("Login failed")
        return False

    def do_login(self, url, user, password, login_form, timeout_s=None):
        self.log_info("Trying to login to '%s' under user %s" % (url, user))
        self.navigate_to(url, cached=None)

        if self._do_login(url, user, password, login_form, timeout_s=timeout_s):
            return True

        for frame in self.dom_find_frames():
            self.dom_switch_to_frame(frame)
            if self._do_login(url, user, password, login_form, timeout_s=timeout_s):
                return True

        self.log_info("Login to '%s' under user '%s' has been failed" % (url, user))
        return False


##############################################################################
# Autotests
##############################################################################


if __name__ == "__main__":
    try:
        b = BrowserWebdriver()
    except NotImplementedError:
        print("OK")
        sys.exit(0)
    sys.exit(-1)



#!/usr/bin/env python

from __future__ import print_function, absolute_import

# -*- coding: utf-8 -*-
__author__ = "perfguru87@gmail.com"
__copyright__ = "Copyright 2018, The PerfTracker project"
__license__ = "MIT"

"""
WordPress admin panel crawler example
"""

import time

from .cp_engine import CPEngineBase, CPMenuItemXpath, CPLoginForm
from .browser_chrome import BrowserChrome


class CPWPAdminConsole(CPEngineBase):
    type = "WordPress admin console"
    menu_dom_clicks = False
    menu_xpaths = [[CPMenuItemXpath(0, None,
                                    "//a[contains(@class, 'menu-top')]",
                                    "./div[contains(@class, 'wp-menu-name')]")],
                   [CPMenuItemXpath(1, None, "//ul[contains(@class, 'wp-submenu')]/li/a", None)]
                   ]

    login_form = CPLoginForm(sbmt_ids=[('input', 'wp-submit')])

    def _get_page_title(self):
        divs = self.browser.driver.find_elements_by_xpath("//div[contains(@id, 'wpbody-content')]/div/h1")
        for div in divs:
            if not div.is_displayed():
                continue
            return div.text
        return ''

    def cp_get_menu_item_title(self, title_el):
        text = title_el.get_attribute("innerHTML")
        return text.split("<")[0]

    def cp_validate_current_page(self, url):
        curr_url = self.cp_get_current_url()

        if "reauth=1" in curr_url:
            self.browser.log_error("WARNING: seems like WP control panel session is expired, trying to re-login")
            time.sleep(2.0)
            if not self.cp_do_login(url):
                raise CPCrawlerException("Re-login failed")
            return False

        return True

    def cp_skip_menu_item(self, link_el, title, link_url):
        if title in ("Customize", "Header", "Editor"):
            return True
        return CPEngineBase.cp_skip_menu_item(self, link_el, title, link_url)

    def cp_menu_item_click(self, el, timeout_s=None, title=None):
        # BackupConsole doesn't re-render the page, it means default
        # dom click wait callback will not work and we need own

        text = self._get_page_title()

        def wait_callback(self, el, timeout_s, title):
            # FIXME: is there a better way to check frame is loaded then to check the title has changed?
            time.sleep(1.0)

            deadline = time.time() + min(timeout_s, 5.0)
            while text != self._get_page_title() and time.time() < deadline:
                time.sleep(1.0)

        self.browser.dom_click(el, timeout_s=timeout_s, name=title, wait_callback=wait_callback, wait_callback_obj=self)


##############################################################################
# Autotests
##############################################################################


if __name__ == "__main__":
    CPWPAdminConsole(BrowserChrome())
    print("OK")



#!/usr/bin/env python

from __future__ import print_function

import sys
from distutils.version import LooseVersion

__version__ = "0.1.2"
__name__ = "perftracker-cp-crawler"

def perftracker_cp_crawler_require_version(ver_required):
    if LooseVersion(__version__) < LooseVersion(ver_required):
        print("Error: perftrackerlib version >= %s must be installed, found %s" %
              (ver_required, __version__))
        sys.exit(-2)



#!/usr/bin/env python

from __future__ import print_function, absolute_import

# -*- coding: utf-8 -*-
__author__ = "perfguru87@gmail.com"
__copyright__ = "Copyright 2018, The PerfTracker project"
__license__ = "MIT"

"""
Naive Python browser, it is able to repeat requests traces collected from a real browser
"""

import os
import sys
import logging
import copy
import socket
import re
import time
import platform
from threading import Lock
from multiprocessing.dummy import Pool as ThreadPool

if sys.version_info[0] < 3:
    from Cookie import SimpleCookie
else:
    from http.cookies import SimpleCookie

from .browser_base import BrowserBase, BrowserExc
from .page import Page, PageRequest, PageRequestsGroup, PageWithActions
from .utils import parse_url, extract_cookies
from .httppool import HTTPPool
from . import httputils


class BrowserPythonNetlocData:
    def __init__(self, browser, netloc):

        # Remove standard port number from netloc string
        # Allowed by HTTP W3C and make some web-apps happy
        if ':' in netloc:
            nl = netloc.split(':')
            if nl[1] == '80' or nl[1] == '443':
                netloc = nl[0]

        self.browser = browser
        self.netloc = netloc

        self.cookies = SimpleCookie()
        self.httpool = {}
        self.lock = Lock()

        self.header = {}
        self.header['User-Agent'] = 'Mozilla/5.0 (X11; Linux x86_64; rv:31.0) Gecko/20100101 Firefox/31.0 PYTHON'
        self.header['Connection'] = 'keep-alive'
        self.header['Accept'] = 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8'
        self.header['Accept-Language'] = 'en-US,en;q=0.5'
        self.header['Content-Type'] = 'application/x-www-form-urlencoded'
        self.header['Host'] = self.netloc

    def _get_http_pool(self, scheme):
        if scheme not in self.httpool:
            loc = "%s://%s" % (scheme, self.netloc)
            self.browser.log_debug("allocating connection pool to %s" % loc)
            self.httpool[scheme] = HTTPPool("%s" % loc)
        return self.httpool[scheme]

    def __del__(self):
        for pool in self.httpool.values():
            pool.clear()

    def set_cookie(self, url, key, val=None, path=None):
        self.browser.log_debug("set cookie: %s, %s=%s" % (url, key, str(val)))

        scheme, netloc, path = parse_url(url)
        with self.lock:
            if type(key) == 'dict':
                c = key
                self.cookies[c['name']] = c
            elif val:
                self.cookies[key] = val
                if path:
                    self.cookies[key]["path"] = path
            elif path in self.cookies:
                del self.cookies[key]

    def set_header(self, key, val):
        self.browser.log_debug("set header: %s=%s" % (key, val))
        with self.lock:
            if val is not None:
                self.header[key] = val
            elif key in self.header:
                del self.header[key]

    def get_cookies_str(self, url):
        """
        get_cookies_str() returns only the cookies applicable to given *url*
        """

        _, _, path = parse_url(url)

        path_cookies = {}
        for key, morsel in self.cookies.items():
            p = morsel["path"] if morsel["path"] else "/"
            if key not in path_cookies:
                path_cookies[key] = {}
            path_cookies[key][p] = morsel.value

        cookies = {}

        for key, paths in path_cookies.items():
            # sort by string length: ["/home/path", "/home", "/"]
            for p in sorted(paths.keys(), lambda x: len(x), reverse=True):
                if path.startswith(p):
                    cookies[key] = paths[p]
                    break

        return "; ".join("%s=%s" % (key, val) for key, val in cookies.items())

    def _get_redirected_request(self, req, response):
        new_req = None
        if req.status in [301, 302, 303]:
            new_req = req.duplicate()
            new_req.method = 'GET'
            new_req.params = None
            new_req.url = response.getheader('location')
        elif self.browser.js_redirects:
            page_actions = httputils.PageActionsExtractor(req.data)
            redirect = page_actions.get_action()
            if redirect:
                new_req = req.duplicate()
                new_req.method = redirect.method
                new_req.params = redirect.get_params()
                new_req.url = redirect.url
                new_req.header['Cookie'] = self.browser.browser_get_cookies_str(req.url)
            req.page_actions = page_actions

        if new_req:
            new_req.validator = None
            new_req.header['Cookie'] = self.browser.browser_get_cookies_str(req.url)
            if not new_req.url.startswith('http'):
                new_req.update_netloc(req.url)
            _, new_req.header['Host'], _ = parse_url(new_req.url)
            return new_req

        return None

    def _execute_page_request(self, pool, page, req, path_with_args):
        assert(req.method == 'POST' or not req.params)  # only POST request may have params

        with pool.borrow() as (conn, is_new):
            # conn.set_debuglevel(1)
            conn.request(req.method, path_with_args, req.params, req.header)
            self.header['Referer'] = req.url
            response = conn.getresponse()
            req.status = response.status
            req.data = response.read()

            self.browser.log_debug(" req %s HTTP response: %s, headers: %s" %
                                   (req.id, req.status, response.getheaders()))

            # handle new cookies
            for cookie in extract_cookies(response):
                if sys.version_info[0] < 3 and isinstance(cookie, unicode):
                    cookie = cookie.encode('ascii', 'ignore')
                self.cookies.load(cookie)

            # handle redirection
            new_req = self._get_redirected_request(req, response)
            if new_req:
                self.browser.log_debug(" req %s, redirect to %s" % (new_req.id, new_req.url))
                page.add_request(new_req)

                # FIXME:
                # 1. _get_netloc_data is private method of the browser.

                self.browser._get_netloc_data(new_req.url).execute_page_request(page, new_req)
                req.status = new_req.status
                req.data = new_req.data

            if req.valid_statuses and req.status not in req.valid_statuses:
                raise BrowserExc(' req %s, %s %s status %d'
                                 (req.id, req.method, req.url, req.status))

        if self.browser.validation:
            req.validate_response(req.data)

        req.complete()
        return

    def execute_page_request(self, page, req):

        scheme, netloc, path_with_args = parse_url(req.url, args=True)

        if scheme not in ("http", "https", "ftp"):
            msg = "Can't execute request on URL with unsupported scheme: %s, %s" % (scheme, netloc)
            self.browser.log_error(msg)
            raise BrowserExc(msg)

        req.start()

        pool = self._get_http_pool(scheme)
        # 10 - is technological retry count to handle closed keep-alive connections
        for loop in range(10):
            try:
                self._execute_page_request(pool, page, req, path_with_args)
                return

            except pool.temporary_errors as ex:
                raise  # aandreev
                # usually it means server has closed keep-alive connection due to timeout. lets retry
                self.browser.log_debug("HTTP Exception: %s %s: %s %s, connection closed? retrying" %
                                       (req.method, req.url, type(ex), str(ex)))
                req.status = str(type(ex))
                time.sleep(0.1 * loop)

            except pool.fatal_errors as ex:
                req.status = str(type(ex))
                req.complete()
                self.browser.log_error("HTTP Exception: %s %s: %s %s" % (req.method, req.url, type(ex), str(ex)))
                return

        req.complete()
        self.browser.log_error("HTTPException: %s %s, all retries failed" % (req.method, req.url))

    def execute_page_request_parallel(self, page, reqs, parallel=8):
        if not len(reqs):
            return
        if len(reqs) == 1:
            self.execute_page_request(page, reqs[0])
            return

        def execute(arg):
            global data
            page, req = arg
            try:
                self.execute_page_request(page, req)
            except RuntimeError:
                import traceback
                self.browser.log_error("traceback:\n" + traceback.format_exc())
            return 0

        args = [(page, req) for req in reqs]

        if parallel > len(reqs):
            parallel = len(reqs)

        pool = ThreadPool(parallel)
        pool.map(execute, args)
        pool.close()
        pool.join()


class BrowserPython(BrowserBase):
    engine = "pybrwsr"

    def __init__(self, headless=True, validation=True, cleanup=True, max_connections=8,
                 js_redirects=False, log_path=None):
        BrowserBase.__init__(self, cleanup=cleanup, log_path=log_path)

        self.validation = validation
        self.max_connections = max_connections
        self.js_redirects = js_redirects  # try to parse page to detect JS and other ways of redirect

        self._netloc_data = {}

    def _get_netloc_data(self, url):
        _, netloc, _ = parse_url(url)
        if netloc not in self._netloc_data:
            self._netloc_data[netloc] = BrowserPythonNetlocData(self, netloc)
        return self._netloc_data[netloc]

    def _browser_navigate(self, location, cached=True, name=None):
        if isinstance(location, Page):
            page = copy.deepcopy(location)
        else:
            return self._http_request("GET", location)

        if not len(page.requests_groups):
            raise BrowserExc("BrowserPython._browser_navigate() - bug: requests group is empty!")

        page.browser = self
        page.start()

        for g in page.requests_groups:
            reqs = g.get_uncached_reqs()
            if not reqs:
                continue
            self._get_netloc_data(page.url).execute_page_request_parallel(page, reqs, parallel=self.max_connections)

        page.complete(self)
        return page

    def _browser_wait(self, page, timeout=None):
        return

    def _browser_warmup_page(self, location, name=None):
        return

    def browser_get_name(self):
        return "Python HTTP/1.x browser"

    def browser_get_version(self):
        return "Python %s" % platform.python_version()

    def browser_set_cookie(self, url, key, val=None, path=None):
        self._get_netloc_data(url).set_cookie(url, key, val, path)

    def browser_get_cookies_str(self, url):
        return self._get_netloc_data(url).get_cookies_str(url)

    def browser_set_header(self, url, key, val):
        self._get_netloc_data(url).set_header(key, val)

    def browser_start(self):
        return os.getpid()

    def browser_stop(self):
        pass

    def browser_reset(self):
        self._netloc_data.clear()

    # === BrowserPython specific === #

    def _http_request(self, method, url, params=None, validator=None, header=None, valid_statuses=None):
        page = Page(self, url)
        req = PageRequest(page)

        req.header.update(self._get_netloc_data(url).header)
        req.header['Cookie'] = self._get_netloc_data(url).get_cookies_str(url)
        if header:
            req.header.update(header)

        req.url = url
        req.method = method
        req.params = params
        req.validator = validator
        req.valid_statuses = valid_statuses

        page.add_request(req)
        page.requests_groups.append(PageRequestsGroup(req))

        ret = self._browser_navigate(page)
        return ret

    def http_get(self, url, params=None, validator=None, header=None, valid_statuses=None):
        return self._http_request("GET", url, params, validator, header, valid_statuses).data

    def http_post(self, url, params=None, validator=None, header=None, valid_statuses=None):
        return self._http_request("POST", url, params, validator, header, valid_statuses).data

    def http_action(self, method, url, params=None, validator=None, header=None, valid_statuses=None):
        # Specialized version: http request executor with extended return
        page = self._http_request(method, url, params, validator, header, valid_statuses)
        request = page.requests[-1]  # get last request in redirected requests chain
        return PageWithActions(request.page_actions, page.data, request.url)

    # === printing facilities === #

    def print_browser_info(self):
        self.print_stats_title("Browser summary")
        print("  - platform: %s" % os.sys.platform)
        print("  - browser:  %s" % self.engine)
        print("  - PID:      %d" % self.pid)


##############################################################################
# Autotests
##############################################################################


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    b = BrowserPython()
    b.navigate_to("https://example.com/")
    b.browser_stop()



#!/usr/bin/env python

from __future__ import print_function, absolute_import

# -*- coding: utf-8 -*-
__author__ = "perfguru87@gmail.com"
__copyright__ = "Copyright 2018, The PerfTracker project"
__license__ = "MIT"

"""
A browser Page helper
"""

import time
import copy
import logging
import sys
import os
import pickle
import base64
from collections import defaultdict

from .utils import parse_url, get_common_url_prefix
from perftrackerlib.helpers.texttable import TextTable


class PageTimeline:
    types = ['navStrt', 'reqStrt', 'rspStrt', 'rspEnd', 'domEnd', 'onloadEnd', 'ajaxEnd']
    jstypes = {'navStrt': 'navigationStart',
               'reqStrt': 'requestStart',
               'rspStrt': 'responseStart',
               'rspEnd': 'responseEnd',
               'domEnd': 'domComplete',
               'onloadEnd': 'loadEventEnd'
               }

    def __init__(self, page, values=None):
        self.page = page
        self.values = {}
        self.deltas = []
        self.ajax_start = 0
        self.total_dur = 0

        for t in self.types:
            self.values[t] = values[t] if values and t in values else 0

        for n in range(0, len(self.types) - 1):
            self.deltas.append(0)

        if page.ts_end and not self.values['ajaxEnd']:
            self.values['ajaxEnd'] = page.ts_end - page.ts_start + self.values['navStrt']

        # fixup

        if self.values['ajaxEnd'] < self.values['onloadEnd']:
            self.values['ajaxEnd'] = self.values['onloadEnd']

        for n in range(0, len(self.types) - 1):
            self.deltas[n] = self.values[self.types[n + 1]] - self.values[self.types[n]]

        self.ajax_start = self.values['domEnd'] - self.values['navStrt']
        self.total_dur = self.values['onloadEnd'] - self.values['navStrt']


class PageEvent:
    def __init__(self, name, params):
        self.name = name
        if type(params) == dict:
            self.params = params
        else:
            try:
                self.params = json.loads(args)
            except ValueError:
                logging.warning("Can't convert event '%s' parameters '%s' to Json" % (name, args))


class PageRequestsGroup:
    def __init__(self, request, threshold_ms=10):
        self.threshold_ms = threshold_ms
        self.ts_start = request.ts_start
        self.ts_end = request.ts_start + request.dur if request.ts_start else None
        self.requests = [request]
        self.uncached_requests = [] if request.cached else [request]

    def _add(self, request):
        if self.ts_start is None:
            self.ts_start = 0
        if self.ts_end is None:
            self.ts_end = 0
        self.ts_start = min(self.ts_start, request.ts_start) if self.ts_start else None
        self.ts_end = max(self.ts_end, request.ts_start + request.dur) if self.ts_end else None
        self.requests.append(request)
        if not request.cached:
            self.uncached_requests.append(request)

    def add_request(self, request):
        if not self.ts_start:
            self._add(request)
            return True
        if self.ts_start <= request.ts_start and request.ts_start <= self.ts_end + self.threshold_ms:
            self._add(request)
            return True
        return False

    def get_uncached_reqs(self):
        return self.uncached_requests


_page_req_id = 0


class PageRequest:
    types = ["Image", "Stylesheet", "Script", "XHR", "Document", "Other"]
    doc_types = ["Document", "Other"]
    types_abbr = {"Image": "IMG", "Stylesheet": "CSS", "Script": "JS", "XHR": "XHR",
                  "Document": "Doc", "Other": "Oth", "Total": "Tot"}

    def __init__(self, page, id=None):
        global _page_req_id
        _page_req_id += 1

        self.page = page
        self.id = id if id else _page_req_id
        self.method = None
        self.url = None
        self.ts_start = None
        self.ts_end = None
        self.content_length = 0
        self.length = 0
        self.dur = 0
        self.connection_reused = False
        self.status = None
        self.type = 'Other'
        self.keepalive = False
        self.gzipped = False
        self.cached = False
        self.completed = False

        self.data = ""
        self.page_actions = None

        # properties below needed to re-execute request
        self.params = None
        self.header = {}
        self.validator = None
        self.valid_statuses = None

        self.longpoll = False

    def __str__(self):
        return str(self.url)

    def __unicode__(self):
        return str(self)

    def duplicate(self):
        result = copy.deepcopy(self)

        global _page_req_id
        _page_req_id += 1
        result.id = _page_req_id

        return result

    def get_url(self, domain):
        if self.url.startswith(domain):
            return self.url[len(domain):]
        return self.url

    def pretty_type(self, type):
        if not type:
            logging.warning("Empty request type for url '%s'" % self.url)
            return "Other"
        tail = self.url.lower()[-6:]
        if tail.endswith('.js'):
            return "Script"
        if tail.endswith('.css'):
            return "Stylesheet"
        if tail.endswith('.png') or tail.endswith('.gif') or tail.endswith('.jpg') or tail.endswith('.ico'):
            return "Image"
        if type in self.types:
            return type
        if type.startswith("text/html") or tail.endswith(".html"):
            return "Document"
        return "Other"

    def set_type(self, type):
        self.type = self.pretty_type(type)

    def start(self, ts=None):
        self.ts_start = ts if ts else int(time.time() * 1000)
        self.page.browser.log_debug(" req %s started   - %d %s %s %s %s" %
                                    (self.id, self.ts_start, self.method, self.url, self.header, self.params))

    def complete(self, ts=None):
        if not ts:
            ts = int(time.time() * 1000)
        self.ts_end = ts
        self.dur = int(round(self.ts_end - self.ts_start))
        self.completed = True
        self.page.browser.log_debug(" req %s completed - %d %s %s %s - %s, %sKA, %sGzip, %sCached,"
                                    " len %d, content-len %d, dur %d ms" %
                                    (self.id, ts, self.method, self.url, self.status, self.type,
                                     "+" if self.keepalive else "-",
                                     "+" if self.gzipped else "-",
                                     "+" if self.cached else "-",
                                     self.length, self.content_length, self.dur))

    def is_long_poll(self, longpolls=None):
        if self.longpoll:
            return True

        if not longpolls:
            return False
        for l in longpolls:
            if l in self.url:
                self.longpoll = True
                return True
        return False

    def is_ok(self):
        if self.longpoll:
            return True
        try:
            base = int(self.status / 100)
        except TypeError:
            return False
        return base in (0, 2, 3)

    def validate_response(self, data):
        if self.validator is not None:
            if data.find(self.validator) < 0:
                data = data.decode('utf-8').encode('ascii', 'ignore')
                from browser import BrowserExc
                raise BrowserExc('%s %s validation failed\nvalidator: %s\ndata: %s' %
                                 (self.method, self.url, self.validator, data))
        return data

    def update_netloc(self, url):
        prot, netloc, _ = parse_url(url)
        self.url = "%s://%s%s" % (prot, netloc, self.url)


class Page:
    def __init__(self, browser, url, cached=True, longpolls=None, name=None, real_navigation=True):
        self.browser = browser
        self.browser_pid = browser.pid if browser else 0  # denormalization required for faster serialization
        self.requests = []  # append in add_request
        self.requests_groups = []  # append on complete()
        self._id2request = {}
        self.longpolls = longpolls

        # FIXME: must be moved to cp_webdriver
        # FIXME: /gelf is a telemtry page on graylog.ap.int.zone/gelf that hangs sometime
        if not self.longpolls:
            self.longpolls = ["notifications?channel=", "notifications/channel", "/gelf", "/api/subscriptions"]

        self.ram_usage_kb = 0
#        self.time_start_utc = int(time.time() * 1000)
        self.ts_start = None
        self.ts_end = None
        self.dur = 0
        self.url = url
        self.name = name
        self.cached = cached
        if url:
            self.domain = parse_url(url, server=True)

        self.length = 0
        self.timeline = PageTimeline(self)
        self.real_navigation = real_navigation  # True - we can trust browser navigation API, false - we can't

        if not self.real_navigation:
            self.cached = True  # FIXME: enforce cached=True for DOM-clicks, otherwise it looks strange in summary

        # page content
        self.data = ""

    def __str__(self):
        return self.url

    def __unicode__(self):
        return self.url

    def __deepcopy__(self, memo):
        # override deepcopy to:
        # 1. avoid 'browser' object cloning
        # 2. reset timelines

        cls = self.__class__
        result = cls(None, self.url, cached=self.cached, longpolls=self.longpolls, name=self.name)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            if k == "browser":
                result.browser = v
            elif k == "timeline":
                result.timeline = None
            else:
                setattr(result, k, copy.deepcopy(v, memo))
        result.ts_start = None
        result.ts_end = None
        result.timeline = PageTimeline(result)
        return result

    def add_request(self, req):
        if req.id in self._id2request:
            self._id2request[req.id] = req
            for n in range(0, len(self.requests)):
                if self.requests[n].id == req.id:
                    self.requests[n] = req
                    return
        else:
            self._id2request[req.id] = req
            self.requests.append(req)

    def del_request(self, req):
        if req.id in self._id2request:
            del self._id2request[req.id]
            for n in range(0, len(self.requests)):
                if self.requests[n].id == req.id:
                    del self.requests[n]
                    return

    def get_request(self, id):
        if id in self._id2request:
            return self._id2request[id]
        return None

    def process_activity(self, name, timestamp):
        if not self.ts_start or self.ts_start > timestamp:
            self.ts_start = timestamp
            self.browser.log_debug("First detected activity '%s' timestamp: %d" % (name, self.ts_start))

        if not self.ts_end or self.ts_end < timestamp:
            self.ts_end = timestamp
            self.browser.log_debug("Last detected activity '%s' timestamp: %d" % (name, self.ts_end))

    def get_incomplete_reqs(self):
        return [r for r in self.requests if not r.completed and not r.is_long_poll(self.longpolls)]

    def get_uncached_reqs(self):
        return [r for r in self.requests if not r.cached]

    def get_error_reqs(self):
        return [r for r in self.requests if not r.is_ok()]

    def get_repeated_reqs_cnt(self):
        urls = [r.url for r in self.get_uncached_reqs()]
        return len(urls) - len(set(urls))

    def get_foreign_reqs(self):
        ret = []
        if not hasattr(self, '__page_netlocs'):
            _, netloc, _ = parse_url(self.url)
            self.__page_netlocs = [netloc, netloc[4:] if netloc.startswith("www.") else "www." + netloc]

        for r in self.get_uncached_reqs():
            _, netloc, _ = parse_url(r.url)
            if netloc not in self.__page_netlocs:
                ret.append(r)
        return ret

    def print_page_requests_stats(self, title=True, description=None):
        print("")
        if title and not isinstance(title, str):
            title = "Navigation's network requests summary"
        if title:
            PageStats.print_title(title)

        PageStats.print_description(description)

        t = TextTable()
        t.add_row(["Type", "Requests", "non-200ok", "non-KA", "non-GZ", "Recv(KB)", "RecvAvg(KB)", "DurAvg(ms)"])
        t.add_row("-")

        uncached = self.get_uncached_reqs()

        for type in PageRequest.types + ["Total"]:
            if type == "Total":
                t.add_row("-")
                items = uncached
            else:
                items = [r for r in uncached if r.type == type]

            if not items:
                continue

            L = sum([r.length for r in items])
            d = sum([r.dur for r in items])
            t.add_row([PageRequest.types_abbr[type],
                       len(items),
                       len([r for r in items if r.status != 200]),
                       len([r for r in items if not r.keepalive]),
                       len([r for r in items if not r.gzipped]),
                       "%8s" % ("%.1f" % (float(L) / 1024)),
                       "%11s" % ("%.1f" % (float(L) / 1024 / len(items))),
                       "%10s" % ("%.0f" % (float(d) / len(items)))
                       ])

        print("  " + "\n  ".join(t.get_lines()))

    def print_page_req_groups_stats(self, title=True, description=None):
        print("")
        if title and not isinstance(title, str):
            title = "Navigation network activity grouped by independent requests"
        if title:
            PageStats.print_title(title)
            print("  Every *group* is separated by the ---- line, group includes independent requests only")
            print("  - Sta   - response HTTP status code")
            print("  - Recv  - length of received data")
            print("  - AX    - is Ajax request")
            print("  - KA    - keep-alive enabled")
            print("  - GZ    - response compressed by gzip\n")

        PageStats.print_description(description)

        t = TextTable(max_col_width=[0, 0, 0, 0, 0, 0, 0, 0, 0, 90], col_separator=" ", left_aligned=[0, 9])
        t.add_row(["Typ", "Sta", " Recv", "AX", "KA", "GZ", "Start", " Dur", "  End", "Url"])
        t.add_row(["", "", " (KB)", "", "", "", " (ms)", "(ms)", " (ms)", ""])

        for g in self.requests_groups:
            reqs = g.get_uncached_reqs()
            if not reqs:
                continue

            t.add_row("-")
            for r in sorted(reqs, key=lambda x: x.ts_start):
                t.add_row(["%s" % PageRequest.types_abbr[r.type],
                           r.status,
                           "%5s" % ("%.1f" % (r.length / 1024.0)),
                           "ax" if (r.ts_start - self.ts_start) > self.timeline.ajax_start else " -",
                           "ka" if r.keepalive else " -",
                           "gz" if r.gzipped else " -",
                           int(round(r.ts_start - self.ts_start)),
                           r.dur,
                           int(round(r.ts_start + r.dur - self.ts_start)),
                           r.get_url(self.domain)])
        print("  " + "\n  ".join(t.get_lines()))

    def start(self, ts=None):
        self.ts_start = ts if ts else int(time.time() * 1000)

    def complete(self, browser, ts=None):
        if not self.ts_end:
            ends = [req.ts_end for req in self.requests]
            self.ts_end = max(ends) if ends else self.timeline.values['ajaxEnd']

        self.ram_usage_kb = browser.browser_get_ram_usage_kb()

        if self.real_navigation:
            self.timeline = browser.browser_get_page_timeline(self)
            if self.timeline.values['navStrt']:
                # Ok, it was real navigation request and we can trust browser navigation API
                if len(self.requests):
                    if not self.ts_start:
                        self.browser.log_error("BUG: page %s load start time is empty" % str(self.url))
                    offt = self.timeline.values['navStrt'] - self.ts_start
                else:
                    offt = 0
                    self.ts_start = self.timeline.values['navStrt']

                self.browser.log_debug("Adjusting page/requests timestamp by %d ms" % round(offt))

                self.ts_start += offt
                self.ts_end += offt
                for r in self.requests:
                    r.ts_start += offt
                    if r.is_long_poll(self.longpolls):
                        self.browser.log_debug("Skip '%s' longpoll request" % r.url)
                        continue
                    elif not r.ts_end:
                        self.browser.log_error("Can't determine '%s' request completion ts, ignored!" % r.url)
                        r.status = "timeout"
                    else:
                        r.ts_end += offt
                for g in self.requests_groups:
                    g.ts_start += offt
                    g.ts_end += offt

        self.length = 0
        for r in self.requests:
            if r.content_length:
                r.length = r.content_length
            if not r.cached:
                self.length += r.length

        self.dur = max(self.ts_end - self.ts_start, self.timeline.total_dur)

        for r in sorted([r for r in self.requests if not r.is_long_poll(self.longpolls)], key=lambda x: x.ts_start):
            if not len(self.requests_groups) or not self.requests_groups[-1].add_request(r):
                self.requests_groups.append(PageRequestsGroup(r))

        # self.data = "".join([req.data for req in self.requests if req.url == self.url])
        for req in self.requests:
            if req.url == self.url:
                if type(req.data) == bytes:
                    self.data += req.data.decode('utf-8')
                else:
                    self.data += req.data

    def get_full_name(self, url_prefix_to_remove=""):
        if self.name:
            return self.name
        return self.url[len(url_prefix_to_remove):]

    def get_key(self, name_priority=False):
        return (self.name if self.name else self.url.split("?bw_id")[0], self.cached)

    def serialize(self, use_pickle=True):
        self.browser.log_debug("serializing page: %s, %d, %s" % (str(int(self.ts_start)), int(self.dur), self.url))
        if use_pickle:
            p = copy.deepcopy(self)
            p.ts_start = self.ts_start  # it was zeroed by deepcopy()
            p.ts_end = self.ts_end  # it was zeroed by deepcopy()
            p.browser = None  # detach real browser object since it has open files
            return base64.urlsafe_b64encode(pickle.dumps(p)) + "\n"

        return "%s|%1.3f|%6d|%2d|%s|%s|%s|%d|%s\n" % \
               (str(int(self.ts_start)), self.dur / 1000.0, self.length, len(self.get_uncached_reqs()),
                "C" if self.cached else "U", self.name if self.name else "", self.browser_pid, str(self.id), self.url)

    @staticmethod
    def deserialize(line, browser_class, use_pickle=True):

        if use_pickle:
            p = pickle.loads(base64.urlsafe_b64decode(line.strip()))
            p.ts_start = int(p.ts_start)
            p.ts_end = int(p.ts_end)
            return p

        ar = line.split("|")
        if len(ar) < 7:
            raise Exception("can't parse line: %s" % line)

        url = ar[7].strip()

        p = Page(None, url)
        p.ts_start = int(ar[0])
        p.dur = float(ar[1].strip()) * 1000.0
        p.ts_end = int(p.ts_start) + int(round(p.dur, 0))
        p.length = int(ar[2])
        p.requests = [] * int(ar[3])
        p.cached = ar[4] == "C"
        p.browser_pid = ar[5].strip()
        p.name = ar[6].strip() if ar[5] else None
        p.url = url

        return p


class PageStatsSummary:
    def __init__(self):
        self.common_prefix = ""
        self.page_stats = []

        def browser_dict():
            return defaultdict(url_dict)

        def url_dict():
            return defaultdict(int)

        def status_dict():
            return defaultdict(int)

        self.reqs = defaultdict(browser_dict)
        self.errs = {}

    def add_page_stats(self, page_stats):
        page_stats.summary = self
        self.page_stats.append(page_stats)

    def update_reqs(self):
        self.reqs = defaultdict(browser_dict)

        urls = set()
        for ps in self.page_stats:
            for p in ps.iterations:
                for r in p.requests:
                    urls.add(r.url)

        self.common_prefix = get_common_url_prefix(urls)

        for ps in self.page_stats:
            for p in ps.iterations:
                for r in p.requests:
                    reqs[ps.id][r.url]['all'] += 1
                    if not r.is_ok():
                        self.reqs[ps.id][r.url][r.status] += 1

        self.errs = {}
        for b, items in self.reqs.items():
            for url, statuses in items.items():
                for status, count in reqs[b][url].items():
                    if status == 'all':
                        continue
                    total = reqs[b][url]['all']
                    if b not in errs:
                        errs[b] = []
                    errs[b].append(["  " + url[len(self.common_prefix):], status, count,
                                   "%.1f" % (100 * count / (1.0 * total))])

    def print_summary(self, title="Summary"):
        print("")
        if title and not isinstance(title, str):
            title = "Summary"
        if title:
            PageStats.print_title(title)

        t = TextTable(left_aligned=[0], max_col_width=[72])
        t.add_row(["Screen", "Iters", "   Requests per page   ", "RecvAvg", "Total", "MemUsg"])
        t.add_row(["", "", "Ntwrk  Rptd  Frgn  Errs", "   (KB)", " (ms)", "  (KB)"])
        t.add_row("-")

        prev_psid = ""
        for ps in self.page_stats:

            if prev_psid != ps.id:
                t.add_row(str(ps.id) + ":")
                prev_psid = ps.id

            t.add_row(["  " + ps.get_screen_title(self.common_prefix), len(ps.iterations),
                       ("%5s  %4s  %4s  %4s") %
                       ("%5.0f" % ps.uncached_reqs if ps.uncached_reqs else "-",
                        "%4.0f" % ps.repeated_reqs if ps.repeated_reqs else "-",
                        "%4.0f" % ps.foreign_reqs if ps.foreign_reqs else "-",
                        "%.1f!" % (ps.errs_cnt) if ps.errs_cnt else "-"),
                       "%.1f" % (ps.size_bytes / 1024.0),
                       "%.0f" % ps.dur_sec,
                       "%.0f" % ps.ram_usage_kb])

        print("  " + "\n  ".join(t.get_lines()))

        if len(self.errs):
            print("")
            PageStats.print_title("Warning: error network requests detected !!!")

            wt = TextTable(left_aligned=[0, 1], max_col_width=[80])
            wt.add_row(["URL", "Status", "Count", "% of total"])
            wt.add_row("-")
            for b, rows in self.errs.items():
                wt.add_row(b + ":")
                for row in rows:
                    wt.add_row(row)
            print("  " + "\n  ".join(wt.get_lines()))


class PageStats:
    separator = "-->"
    width = 118

    def __init__(self, id=""):
        self.iterations = []
        self.id = id
        self.summary = None

        self.update()

    def update(self):
        self.size_bytes = 0
        self.errs_cnt = 0
        self.uncached_reqs = 0
        self.repeated_reqs = 0
        self.foreign_reqs = 0
        self.dur_sec = 0
        self.ram_usage_kb = 0

        if not len(self.iterations):
            return

        for i in self.iterations:
            self.size_bytes += i.length
            self.errs_cnt += len(i.get_error_reqs())
            self.uncached_reqs += len(i.get_uncached_reqs())
            self.repeated_reqs += i.get_repeated_reqs_cnt()
            self.foreign_reqs += len(i.get_foreign_reqs())
            self.dur_sec += i.dur
            self.ram_usage_kb += i.ram_usage_kb

        n = float(len(self.iterations))

        self.size_bytes /= n
        self.errs_cnt /= n
        self.uncached_reqs /= n
        self.repeated_reqs /= n
        self.foreign_reqs /= n
        self.dur_sec /= n
        self.ram_usage_kb /= n

    def add_iteration(self, page):
        self.iterations.append(page)
        self.update()
        if self.summary:
            self.summary.update_reqs()

    @staticmethod
    def print_title(title):
        print(title.upper())
        print("=" * len(title))
        print("")

    @staticmethod
    def print_description(description):
        if description:
            print("  " + "\n  ".join(description))
            print("")

    @staticmethod
    def print_page_timeline_header(title=True, description=None):
        print("")
        if title and not isinstance(title, str):
            title = "Page(s) timeline and memory usage"
        if title:
            PageStats.print_title(title)

            j = PageTimeline.jstypes
            print("  http://www.w3.org/TR/navigation-timing/timing-overview.png")
            print("  http://www.w3.org/TR/navigation-timing/#sec-navigation-timing-interface")
            print("  - navStrt    - %s - navigation start" % j['navStrt'])
            print("  - reqStrt    - %s - browser has resolved the domain (and redirects)"
                  " and started the request" % j['reqStrt'])
            print("  - rspStrt    - %s - server started response (TTFB)" % j['rspStrt'])
            print("  - rspEnd     - %s - browser received last byte of the document" % j['rspEnd'])
            print("  - domEnd     - %s - browser has downloaded all the CSS, JS and rendered the DOM model" %
                  j['domEnd'])
            print("  - onloadEnd  - %s - browser completed the onload() callbacks" % j['onloadEnd'])
            print("  - ajaxEnd    - last detected browser activity, all pending img/css/js and Ajax requests"
                  " have been completed")
            print("  - Total(ms)  - total time of page (TTLB) including all the img/css/js and Ajax requests")
            print("  - MemUsg(KB) - page memory usage (RSS delta between browser start and after page fully loaded)")
            print("")

        PageStats.print_description(description)

        print("  Iter # " + " |", end=" ")
        for n in range(0, len(PageTimeline.types) - 1):
            print("%s %s" % (PageTimeline.types[n], PageStats.separator), end=" ")
        print(PageTimeline.types[-1], end=" ")
        print("| Total(ms) | MemUsg(KB)")
        print("  " + "-" * PageStats.width)

    def print_page_timeline(self, p, title="", hr=False):
        if not p:
            return

        if hr:
            print("  " + "-" * self.width)

        print("  %7s |" % title, end=" ")
        t = p.timeline
        for n in range(0, len(t.deltas)):
            print("%s%5d" % (" " * (len(PageTimeline.types[n]) - 4 + len(self.separator)), t.deltas[n]), end=" ")
        print("%s" % (" " * len(PageTimeline.types[-1])), end=" ")
        print("| %9d | %10d" % (p.dur, p.ram_usage_kb))

    def get_avg(self, iterations=None):
        if not iterations:
            iterations = self.iterations
        if not iterations or len(iterations) < 2:
            return None

        avg = Page(None, "", None)
        avg.iterations = len(iterations)
        for p in iterations:
            t = p.timeline
            for d in range(0, len(t.deltas)):
                avg.timeline.deltas[d] += t.deltas[d]
            avg.dur += p.dur

        for d in range(0, len(t.deltas)):
            avg.timeline.deltas[d] = int(avg.timeline.deltas[d] / len(iterations))
        avg.dur = int(avg.dur / len(iterations))
        avg.ram_usage_kb = sum([p.ram_usage_kb for p in iterations]) / len(iterations)
        return avg

    def get_screen_title(self, common_prefix=""):
        if len(self.iterations):
            return self.iterations[0].get_full_name(common_prefix)
        return ""


# Represents some kind of the html page model with focus on actions (URLs)
class PageWithActions:
    def __init__(self, actions, body, url):
        self.actions = actions  # actions (urls) found on page
        self.body = body  # page body (html text)
        self.url = url  # page source url



#!/usr/bin/env python

from __future__ import print_function, absolute_import

# -*- coding: utf-8 -*-
__author__ = "perfguru87@gmail.com"
__copyright__ = "Copyright 2018, The PerfTracker project"
__license__ = "MIT"

"""
Any control panel engine (helper)
"""

import sys
import os
import logging
import time
import re

from .browser_base import BrowserBase, BrowserExc, BrowserExcTimeout
from .browser_python import BrowserPython
from selenium.common.exceptions import ElementNotVisibleException
from selenium.common.exceptions import NoSuchElementException
from selenium.common.exceptions import WebDriverException
from selenium.common.exceptions import StaleElementReferenceException


reHTML = re.compile('<.*?>')


def remove_html_tags(text):
    return re.sub(reHTML, '', text)


class CPMenuItemXpath:
    def __init__(self, level, frame, link_xpath, title_xpath, menu_url_clicks=True, menu_dom_clicks=True):
        self.level = level  # menu level, 0 - 10...
        self.frame = frame  # menu frame name
        self.link_xpath = link_xpath  # clickable menu item element xpath
        self.title_xpath = title_xpath  # relative xpath to fetch the menu item title
        self.menu_url_clicks = menu_url_clicks
        self.menu_dom_clicks = menu_dom_clicks


class CPMenuItem:
    def __init__(self, level, title, link, xpath, parent, menu_url_clicks=True, menu_dom_clicks=True):
        """
        menu_url_clicks - collect direct URL links to menu items
        menu_dom_clicks - collect DOM (xpath) links to menu items
        """
        self.level = level
        self.link = link
        self.xpath = xpath
        self.children = []
        self._scanned_menu_items = set()

        self.parent = parent
        if parent:
            self.title = parent.title + " -> " + title
        else:
            self.title = title
        self.menu_url_clicks = menu_url_clicks
        self.menu_dom_clicks = menu_dom_clicks

        if link:
            print("  %s - %s" % (self.title, link))  # ugly :-(

    def is_scanned(self, key, check_in_parent=False):
        if check_in_parent and self.parent:
            return self.parent.is_scanned(key, check_in_parent=check_in_parent)
        return key in self._scanned_menu_items

    def mark_as_scanned(self, key):
        p = self
        while p.parent:
            p = p.parent
        p._scanned_menu_items.add(key)
        self._scanned_menu_items.add(key)

    def add_child(self, title, link, xpath, menu_xpath):
        ch = CPMenuItem(self.level + 1, title, link, xpath, self,
                        menu_url_clicks=self.menu_url_clicks and menu_xpath.menu_url_clicks,
                        menu_dom_clicks=self.menu_dom_clicks and menu_xpath.menu_dom_clicks,
                        )
        self.children.append(ch)

        self.mark_as_scanned(title)
        self.mark_as_scanned(link)
        return ch

    def get_items(self, items=None):
        """ return a list of [item#, title, link] """
        if not items:
            items = {}
        for c in self.children:
            if self.menu_url_clicks:
                items[c.link] = [len(items), c.title, c.link]
            if self.menu_dom_clicks and c.xpath:
                xpath = "%s^%s" % (c.link, c.xpath)
                items[xpath] = [len(items), c.title + " (DOM click)", xpath]
            items = c.get_items(items)
        return items


class CPLoginForm:
    def __init__(self, user_tags=None, user_ids=None, pass_tags=None, pass_ids=None,
                 sbmt_tags=None, sbmt_ids=None, sbmt_xpath=None):

        def _capitalize_list(tags):
            ret = [k for k in tags]
            for k in tags:
                if type(k) is tuple:
                    ret.append((k[0].capitalize(), k[1].capitalize()))
                else:
                    ret.append(k.capitalize())
            return ret

        if user_tags is None:
            user_tags = _capitalize_list([("input", "user"), ("input", "username"), ("input", "login")])
        if user_ids is None:
            user_ids = _capitalize_list([("input", "user_login")])
        if pass_tags is None:
            pass_tags = _capitalize_list([("input", "pass"), ("input", "password"), ("input", "login_password")])
        if pass_ids is None:
            pass_ids = _capitalize_list([("input", "user_pass")])
        if sbmt_tags is None:
            sbmt_tags = _capitalize_list([("button", "login"), ("button", "login_submit")])
        if sbmt_ids is None:
            sbmt_ids = _capitalize_list([("button", "login"), ("button", "submit"), ("button", "signin_button")])

        self.user_tags = user_tags
        self.user_ids = user_ids
        self.pass_tags = pass_tags
        self.pass_ids = pass_ids
        self.sbmt_tags = sbmt_tags
        self.sbmt_ids = sbmt_ids
        self.sbmt_xpath = sbmt_xpath if sbmt_xpath else []

        assert isinstance(self.sbmt_xpath, list)


class CPEngineBase:
    type = "A control panel"
    menu_url_clicks = True  # collect direct URL links to menu items
    menu_dom_clicks = True  # collect DOM (xpath) links to menu items
    menu_xpaths = []  # [[CPMenuItemXpath(0, ...), ...], [CPMenuItemXpath(1, ...), ...]]

    login_form = CPLoginForm()

    def __init__(self, browser, user=None, password=None, html_report=None):
        self.browser = browser
        self.log_error = browser.log_error
        self.log_warning = browser.log_warning
        self.log_info = browser.log_info
        self.log_debug = browser.log_debug
        self.menu = CPMenuItem(0, self.type, None, None, None,
                               menu_url_clicks=self.menu_url_clicks, menu_dom_clicks=self.menu_dom_clicks)
        self.current_frame = None
        self.user = user
        self.password = password
        self._html_report = html_report

    #
    # Virtual methods, can be control panel specific
    #

    def cp_init_context(self):
        return True

    def cp_handle_opts(self, opts):
        return

    def cp_validate_current_page(self, url):
        return True

    def cp_get_product_version(self):
        return None

    def cp_get_product_name(self):
        return None

    def cp_get_current_url(self, url=None):
        if url and url.lower().find('javascript') < 0:
            return url
        return self.browser.browser_get_current_url()

    def cp_restore_nav_menu_url(self):
        return False

    def cp_get_menu_item_title(self, title_el):
        return remove_html_tags(title_el.get_attribute("innerHTML")).strip()

    def cp_get_menu_item_link_url(self, link_el):
        return ""

    def cp_skip_menu_item(self, link_el, title, url):
        return not link_el.is_displayed()

    def cp_do_menu_item_click(self, el, timeout_s=None, title=None):
        def wait_callback(self, el, timeout_s, title):
            self.browser.dom_wait_element_stale(el, timeout_s=timeout_s, name=title)

        self.browser.dom_click(el, timeout_s=timeout_s, name=title,
                               wait_callback=wait_callback, wait_callback_obj=self)

    def cp_do_scan_menu(self):
        self.browser.print_stats_title("Control panel menu scanner...")
        print("Control panel detected: '%s'" % self.type)  # ugly :-(
        print("Searching for menu items...\n")  # ugly :-(
        self.menu.link = self.cp_get_current_url()
        self._populate_menu(self.menu)
        self.log_info("Menu scan completed")
        items = self.menu.get_items()
        return [(i[1], i[2]) for i in sorted(items.values(), key=lambda x: x[0])]

    def cp_do_navigate(self, location, timeout=None, cached=True, stats=True, name=None):
        for retry in range(0, 3):
            page = self.browser.navigate_to(location, timeout=timeout, cached=cached, stats=stats, name=name)
            if self.cp_validate_current_page(location):
                return page
            time.sleep(1.0)

        self.log_error("Page url validation failed, aborting")
        return None

    def cp_do_login(self, url, timeout_s=None):
        return self.browser.do_login(url, self.user, self.password, self.login_form, timeout_s=timeout_s)

    def cp_do_logout(self):
        return False

    #
    # Internal / private methods
    #

    def get_current_xpath(self, link_el):
        if not link_el:
            return None

        xpath = "a"

        parent = link_el
        tag = "a"

        while True:
            parent = parent.find_element_by_xpath("..")
            if not parent:
                break

            id = parent.get_attribute('id')
            tag = parent.tag_name

            if id:
                xpath = "%s[@id='%s']/%s" % (tag, id, xpath)
                break

            xpath = "%s/%s" % (tag, xpath)

        xpath = "//%s" % xpath
        return xpath

    def switch_to_frame(self, frame, verbose=True):
        if not frame:
            return None
        self.switch_to_default_content()

        self.browser.log_info("searching for frame: '%s'" % frame)
        try:
            el = self.browser.driver.find_element_by_xpath(frame)
        except NoSuchElementException:
            if verbose:
                self.browser.log_error("Can't find frame element: '%s', page source:\n%s" %
                                       (frame, self.browser.driver.page_source))
            return None
        self.browser.dom_switch_to_frame(el)
        self.current_frame = frame
        return el

    def switch_to_default_content(self):
        self.current_frame = None
        self.browser.dom_switch_to_default_content()

    def dom_click(self, el, title=None):
        self.browser.dom_click(el, name=title)

    def _populate_menu(self, menu):

        if len(self.menu_xpaths) <= menu.level:
            return

        if menu.link != self.cp_get_current_url():
            if self.cp_restore_nav_menu_url():
                self.log_debug("Restore navigatio menu URL back to: '%s'" % menu.link)
                self.browser.navigate_to(menu.link)

        for x in self.menu_xpaths[menu.level]:

            if x.frame:
                frame = self.switch_to_frame(x.frame, verbose=False)
                if not frame:
                    continue

            self.log_debug("Looking for xpath: '%s'" % x.link_xpath)
            menu_elements = self.browser.driver.find_elements_by_xpath(x.link_xpath)
            i = 0
            while i < len(menu_elements):
                link_el = menu_elements[i]

                link_url = self.cp_get_menu_item_link_url(link_el)

                if not link_url:
                    try:
                        link_url = link_el.get_attribute('href')
                    except StaleElementReferenceException:
                        # previous click caused dom change, so re-load menu items (assuming their sequence is preserved)
                        menu_elements = self.browser.driver.find_elements_by_xpath(x.link_xpath)
                        if i >= len(menu_elements):
                            continue  # menu has shrunk suddenly
                        link_el = menu_elements[i]
                        link_url = link_el.get_attribute('href')
                    i += 1

                if not link_url:
                    link_url = link_el.get_attribute('innerHTML').strip()
                    link_url = remove_html_tags(link_url)
                    if "/" not in link_url and "#" not in link_url:
                        link_url = ""  # the link is not known at the moment, can't do nothing

                if x.title_xpath:
                    title_els = link_el.find_elements_by_xpath(x.title_xpath)
                    if not title_els or not len(title_els) or not title_els[0].get_attribute("innerHTML"):
                        self.log_error("WARNING: can't get title for menu element: %s\nusing: %s" %
                                       (link_url, x.title_xpath))
                        continue
                    title_el = title_els[0]
                else:
                    title_el = link_el

                title = self.cp_get_menu_item_title(title_el)

                msg = "menu item: '%s' => '%s'" % (title, link_url)

                if menu.is_scanned(title):
                    self.log_info("skip %s [already scanned]" % msg)
                    continue

                if link_url == "javascript:void(0);":
                    self.log_info("skip %s [void js link]" % msg)
                    continue

                if self.cp_skip_menu_item(link_el, title, link_url):
                    self.log_info("skip %s [cp_skip_menu_item() = True]" % msg)
                    continue

                curr_xpath = self.get_current_xpath(link_el)

                self.log_info("found %s" % msg)
                try:
                    self.cp_do_menu_item_click(link_el, title=title)
                except WebDriverException as e:
                    self.log_info(" ... skipping the '%s' menu item: %s" % (title, str(e)))
                except ElementNotVisibleException:
                    self.log_info(" ... skipping the '%s' menu item since it is not visible" % title)
                    continue
                except BrowserExc as e:
                    self.log_debug("dom_click() raised exception: %s" % str(e))
                    self.log_warning("WARNING: can't wait for '%s' " % (title) + "menu click completion, skipping it")
                    continue

                curr_url = self.cp_get_current_url(link_url)
                self.browser.history.append(curr_url)

                if menu.is_scanned(curr_url, check_in_parent=True):
                    self.log_debug("skipping menu item with url '%s', it was already scanned" % curr_url)
                    continue

                self.log_info(" ... '%s' menu item has URL: %s, xpath: %s" % (title, curr_url, curr_xpath))

                ch = menu.add_child(title, curr_url, curr_xpath, x)

                if self._html_report:
                    img_path = self._html_report.add_page(url=curr_url, title=ch.title)
                    self.browser.browser_get_screenshot_as_file(img_path)
                    self._html_report.gen_thumbnails(curr_url)

                self._populate_menu(ch)
                self._populate_menu(menu)
                break

            if x.frame:
                self.switch_to_default_content()


##############################################################################
# Autotests
##############################################################################


if __name__ == "__main__":
    CPEngineBase(BrowserPython())
    print("OK")



#!/usr/bin/env python

from __future__ import print_function, absolute_import

# -*- coding: utf-8 -*-
__author__ = "perfguru87@gmail.com"
__copyright__ = "Copyright 2018, The PerfTracker project"
__license__ = "MIT"

"""
Base browser library
"""

import sys
import time
import tempfile
import re
import logging
import os
import threading
import traceback
import datetime
import platform

if sys.version_info[0] < 3:
    import httplib
else:
    import http.client as httplib

# time to wait for the incomplete requests
DEFAULT_NAV_TIMEOUT = 60.0

# time to wait for browser ajax activity (in seconds) to consider page load is fully complete
DEFAULT_AJAX_THRESHOLD = 2.0

"""
TODO:
- Network connection simulation (3G, LTE, etc) (Use tc on linux, pfctl on Mac ?)
- IE support (needed for Windows)
- handle 'Server not found'
"""

"""
Glossary

Browser classes:

  BrowserBase class    - base browser class representing API available to the library users
  |- BrowserWebdriver  - base class for webdriver (i.e. selenium) based browsers (Chrome, Firefox)
  |  |- BrowserChrome   - Chrome browser
  |  `- FirefoxBrowser  - Firefox browser
  `- BrowserPython      - Python browser
     `- CPBase          - Control Panel base class

Page classes:

  PageStats             - class collecting all the page request to given URL
  Page                  - class describing a page request and response (i.e. navigation item in the prowser)
  PageRequest           - class describing individual sub-requests of a page (css, js, html, etc)
  PageTimeline          - represents page timeline
  PageWithActions       - represents some kind of the html page model with focus on actions (URLs)

"""

from .page import Page, PageStats, PageTimeline


################################################################
# Exceptions
################################################################


class BrowserExc(Exception):
    def __init__(self, msg=""):
        Exception.__init__(self, "%s" % msg)
        self.message = "browser module exception: %s" % msg


class BrowserExcTimeout(BrowserExc):
    def __init__(self, msg=""):
        BrowserExc.__init__(self, "%s" % msg)
        self.message = "browser module timeout: %s" % msg


class BrowserExcNotImplemented(BrowserExc):
    pass


class BrowserExcNotSupported(BrowserExc):
    """
    Raise this exception if for instance given method is intentionally not supported in a class
    """
    def __init__(self):
        BrowserExc.__init__(self)
        tb = traceback.extract_tb()  # FIXME
        self.message += "method %s is not supported in %s" % (str(tb), "FIXME Class")


class BrowserExcError(BrowserExc):
    def __init__(self, message):
        BrowserExc.__init__(self)
        self.message += message


################################################################
# Browser class
################################################################

_browser_id = {}


class BrowserBase:
    engine = 'basebrwsr'

    def __init__(self, headless=True, resolution=(1440, 900), cleanup=True, telemetry_fname=None,
                 log_path='auto', nav_timeout=DEFAULT_NAV_TIMEOUT, ajax_threshold=DEFAULT_AJAX_THRESHOLD,
                 remote_connstring=None):
        self.history = []
        self.page_stats = {}

        self._name = self._init_name()

        self.resolution = resolution
        self.driver = None
        self.display = None
        self.pid = None
        self.nav_timeout = nav_timeout
        self.ajax_threshold = ajax_threshold
        self.remote_connstring = remote_connstring

        if log_path == 'auto':
            self.log_path = tempfile.NamedTemporaryFile(delete=cleanup).name
        else:
            self.log_path = log_path

        if telemetry_fname:
            self.telemetry_log = open(telemetry_fname, 'a', 0)
        else:
            self.telemetry_log = None

        if self.log_path:
            self.log_info("log path: %s" % self.log_path)

        self._browser_display_init(headless, resolution)
        self.pid = self.browser_start()
        if not self.pid:
            raise BrowserExc("Browser initialization failed")

        self._base_rss_kb = self._browser_get_rss_kb()

        # looping

        self._loop_locations = []
        self._loop_stop = True
        self._loop_thread = None
        self._loop_sleep_sec = 0

    def _init_name(self):
        global _browser_id
        id = _browser_id.get(self.engine, 0)
        _name = "%s#%02d" % (self.engine, id)
        _browser_id[self.engine] = id + 1
        return _name

    def __del__(self):
        self.browser_stop()

    # === browser* methods ===#

    def _browser_clear_caches(self):
        """
        clear browser caches
        """
        self.history = []

    def _browser_navigate(self, location, cached=True, name=None):
        raise NotImplementedError

    def _browser_wait(self, page, timeout=None):
        """
        wait for navigation request completion
        """
        raise NotImplementedError

    def _browser_warmup_page(self, location, name=None):
        """
        warmup (i.e. cache) a page
        """
        raise NotImplementedError

    def _browser_display_init(self, headless, resolution):
        return

    def _browser_get_current_url(self):
        raise NotImplementedError

    def browser_get_name(self):
        raise NotImplementedError

    def browser_get_version(self):
        raise NotImplementedError

    def browser_get_platform(self):
        raise platform.platform()

    def _browser_get_rss_kb(self):
        return 0

    def browser_get_ram_usage_kb(self):
        return self._browser_get_rss_kb() - self._base_rss_kb

    def browser_get_screenshot_as_file(self, filename):
        raise NotImplementedError

    def browser_get_page_timeline(self, page):
        return PageTimeline(page)

    def browser_start(self):
        raise NotImplementedError

    def browser_stop(self):
        """
        exit from the driver
        """
        raise NotImplementedError

    def browser_reset(self):
        """Reset cookies and connection pools"""
        pass

    # === domain methods === #

    def domain_get_cookies(self, url):
        raise NotImplementedError

    def domain_set_cookies(self, url, cookies):
        raise NotImplementedError

    def domain_set_cookie(self, url, key, val=None, path=None):
        raise NotImplementedError

    def domain_set_header(self, url, key, val):
        raise NotImplementedError

    def domain_set_session(self, url, session_id):
        raise NotImplementedError

    # ==== logging === #

    def log_debug(self, msg):
        logging.debug(msg, extra={'browser': self._name})

    def log_info(self, msg):
        logging.info(msg, extra={'browser': self._name})

    def log_warning(self, msg):
        logging.warning(msg, extra={'browser': self._name})

    def log_error(self, msg):
        logging.error(msg, extra={'browser': self._name})

    def event_log(self, p):
        if not self.telemetry_log:
            return
        self.telemetry_log.write(p.serialize())

    # === Navigation looping === #

    def _loop(self):
        try:
            while not self._loop_stop:
                for loc in self._loop_locations:
                    self.navigate_to(loc)
                time.sleep(self._loop_sleep_sec)

        except httplib.BadStatusLine as e:
            pass
        except RuntimeError:
            logging.error("traceback:\n" + traceback.format_exc())

    def loop_start(self, locations, sleep_sec=0):
        self.log_debug("loop_start():\n   %s" % "\n   ".join([str(loc) for loc in locations]))
        if self._loop_thread:
            self.loop_stop()

        self._loop_locations = locations
        self._loop_sleep_sec = sleep_sec
        self._loop_stop = False
        self._loop_thread = threading.Thread(target=self._loop)
        self._loop_thread.start()

    def loop_stop(self):
        self.log_debug("loop_stop()")
        self._loop_stop = True

    def loop_wait(self):
        self.log_debug("loop_wait()")
        self._loop_thread.join()
        self._loop_thread = None

    # === navigation API === #

    def navigate_to(self, location, timeout=None, cached=True, stats=True, name=None):
        """
        navigate to given url or page in cached/uncached mode
        """
        url = location.url if isinstance(location, Page) else location

        scheme = url.split(":")[0]

        if scheme not in ("http", "https", "file", "about", "chrome"):
            url = "http://%s" % url

        if isinstance(location, Page):
            location.url = url
        else:
            location = url

        if cached is None:
            self.log_info("Navigate to: %s" % url)
        else:
            self.log_info("Navigate to: %s %s" % (url, "CACHED" if cached else "UNCACHED"))

            if cached and url not in self.history:
                self._browser_warmup_page(location, name=name)
            elif not cached and url in self.history:
                self._browser_clear_caches()

        p = self._browser_navigate(location, cached=cached, name=name)
        self._browser_wait(p, timeout=timeout)

        self.history.append(url)

        if stats:
            key = p.get_key()
            if key not in self.page_stats:
                self.page_stats[key] = PageStats(len(self.page_stats))
            self.page_stats[key].add_iteration(p)

        self.event_log(p)
        self.log_info("Navigation completed: %s %s, dur %d ms" % (url, "CACHED" if cached else "UNCACHED", p.dur))
        return p

    def navigation_reset(self):
        pass

    # === printing facilities === #

    @staticmethod
    def print_stats_title(title):
        print("")
        print(title.upper())
        print("=" * len(title))
        print("")

    def print_browser_info(self):
        raise NotImplementedError


##############################################################################
# Autotests
##############################################################################


if __name__ == "__main__":
    try:
        b = BrowserBase()
    except NotImplementedError:
        print("OK")
        sys.exit(0)
    sys.exit(-1)



#!/usr/bin/env python

from __future__ import print_function, absolute_import

# -*- coding: utf-8 -*-
__author__ = "perfguru87@gmail.com"
__copyright__ = "Copyright 2018, The PerfTracker project"
__license__ = "MIT"

"""
A library to scan a control panel menu items, click on them, wait for request completion and report
performance metrics.

Features:
- Native Chrome and Firefox browsers support
- Automatic menu and sub-menu discovery (by xpath patterns)
- Login form detection and authentication
- Run multiple native browsers in background under different users
- Automatic ajax activity detection in browser and wait for click completion
- Python-based browser repeaters to simulate bigger load
- Waterfall-like requests view
- Page click requests errors tracking
- Cached/Uncached clicks/pages
- Direct URL click / xpath click
- Headless / traditional browsers
"""

import os
import sys
import logging
import traceback
import time
import re
import shutil
import copy
import datetime
import socket
import platform
import math
from tempfile import gettempdir
from optparse import OptionParser, OptionGroup
from multiprocessing import Process, Queue, cpu_count
from psutil import virtual_memory

from perftrackerlib import __version__ as __version__
from perftrackerlib.client import ptSuite, ptTest, ptVM, ptComponent
from .browser_base import BrowserExc, DEFAULT_NAV_TIMEOUT, DEFAULT_AJAX_THRESHOLD
from .browser_python import BrowserPython
from .browser_chrome import BrowserChrome
from .browser_firefox import BrowserFirefox
from .html_report import ptBrowserHtmlReport
from .page import PageStats, PageStatsSummary
from .utils import gen_urls_from_index_file
from .cp_engine import CPEngineBase
from perftrackerlib.helpers.texttable import TextTable
from perftrackerlib.helpers.ptshell import ptShell
from selenium.webdriver.remote.remote_connection import LOGGER as selenium_logger

bindir, basename = os.path.split(sys.argv[0])
basename = basename.split(".")[0]

BROWSERS = (BrowserChrome, BrowserFirefox, BrowserPython)

OPT_DELAY_BETWEEN_CLICK_SEC = 1.0
OPT_DEFAULT_AJAX_THRESHOLD = DEFAULT_AJAX_THRESHOLD
OPT_DEFAULT_NAV_TIMEOUT = DEFAULT_NAV_TIMEOUT


class CPLogFormatter(logging.Formatter):
    LOG_LVL_MAP = {logging.CRITICAL: "CRI",
                   logging.ERROR: "ERR",
                   logging.WARNING: "WRN",
                   logging.INFO: "INF",
                   logging.DEBUG: "DBG",
                   logging.NOTSET: "n/a"
                   }

    def __init__(self, *args, **kwargs):
        return super(CPLogFormatter, self).__init__("%(message)s", *args, **kwargs)

    def format(self, record):
        if 'browser' in record.__dict__:
            name = record.__dict__['browser']
        else:
            name = record.name[0:10].split(".")[0]
        record.msg = '%s [%3s] %-10s - %s' % (datetime.datetime.now().strftime("%H:%M:%S.%f")[:-3],
                                              self.LOG_LVL_MAP[record.levelno], name, record.msg)
        return super(CPLogFormatter, self).format(record)


class CPCrawlerException(RuntimeError):
    pass


class CPBrowserRunner:
    def __init__(self, cp_engines, opts, urls, users, browser_id, logfile, workdir, pt_suite):
        self.cp_engines = cp_engines
        self.opts = opts
        self.urls = urls
        self.browser_id = int(browser_id)
        self.workdir = workdir
        self.page_stats_summary = PageStatsSummary()
        self.user = users[(self.browser_id - 1) % len(users)] if users else None
        self.pt_suite = pt_suite
        self._html_report = None

        self.browser_class = BROWSERS[0]
        for b in BROWSERS:
            if b.engine == self.opts.browser:
                self.browser_class = b
                break

        self.logdir = os.path.join(workdir, "browser.%d" % browser_id)
        self.crawler_logfile = os.path.join(self.logdir, logfile if logfile else "%s.log" % basename)
        self.browser_logfile = os.path.join(self.logdir, "%s.log" % self.browser_class.engine)

        self.stdout_fname = os.path.join(self.logdir, "%s.stdout" % basename) if self.browser_id else None
        self.stderr_fname = os.path.join(self.logdir, "%s.stderr" % basename) if self.browser_id else None

        self._stdout_orig = sys.stdout
        self._stderr_orig = sys.stderr

        if self.browser_id:
            self.opts.log2file = True

    def init(self):
        shutil.rmtree(self.logdir, ignore_errors=True)
        os.makedirs(self.logdir, mode=0o777)

        self.browser = self.browser_class(headless=not self.opts.view, cleanup=False,
                                          telemetry_fname=self.opts.telemetry,
                                          log_path=self.browser_logfile,
                                          nav_timeout=self.opts.nav_timeout,
                                          ajax_threshold=self.opts.ajax_threshold,
                                          remote_connstring=self.opts.remote_connstring)

        if self.browser_id:
            sys.stdout = open(self.stdout_fname, 'w')
            sys.stderr = open(self.stderr_fname, 'w')

            self.opts.telemetry = os.path.join(self.workdir, "telemetry.log")

        if self.opts.html_report and self.browser_id == 0:
            self._html_report = ptBrowserHtmlReport(self.opts.html_report, title=self.urls[0])
            self._html_report.gen_index_html()

    def fini(self):
        if self.browser_id:
            sys.stdout = self._stdout_orig
            sys.stderr = self._stderr_orig

        if self._html_report:
            p = self._html_report.gen_index_html()
            print("See HTML report here: %s" % p)

    def _detect_cp_type(self):
        for cp in self.cp_engines:
            self.browser.log_info("Trying to recognize: %s ..." % cp.type)
            c = cp(self.browser, self.user, self.opts.password, self._html_report)
            if not c.cp_init_context():
                continue
            if not len(c.menu_xpaths):
                continue
            for cpmx in c.menu_xpaths[0]:
                c.switch_to_frame(cpmx.frame, verbose=False)
                if self.browser.driver.find_elements_by_xpath(cpmx.link_xpath):
                    self.browser.log_info("%s control panel detected" % c.type)
                    c.switch_to_default_content()
                    return c
                c.switch_to_default_content()
        return None

    def _login(self, url, wait_completion=True):
        self.browser.navigate_to(url, cached=False)
        for cp in self.cp_engines:
            self.browser.log_info("Trying to login into: %s ..." % cp.type)
            c = cp(self.browser, self.user, self.opts.password)
            if not c.cp_init_context():
                continue
            if c.cp_do_login(url, timeout_s=None if wait_completion else 2.0):
                return True
        logging.error("Login to %s under %s:%s failed" % (url, self.user, self.opts.password))
        sys.exit(-1)

    def _pt_suite_init(self, cp):
        if not self.opts.pt_project:
            return

        suite_ver = self.opts.pt_version if self.opts.pt_version else __version__
        suite_name = "cp_crawler"
        product_ver = self.opts.pt_product_version if self.opts.pt_product_version else cp.cp_get_product_version()
        product_name = self.opts.pt_product_name if self.opts.pt_product_name else cp.cp_get_product_name()

        self.pt_suite.suite_ver = suite_ver
        self.pt_suite.suite_name = suite_name
        self.pt_suite.product_ver = product_ver
        self.pt_suite.product_name = product_name

        if not self.opts.pt_title and product_name and product_ver:
            browser_name = self.browser.browser_get_name()
            browser_name += self.browser.browser_get_version().split(".")[0]
            ghz = ptShell().hw_info.cpu_freq_ghz
            self.pt_suite.job_title = "%s %s @ %s %1.fGHz" % (product_name, product_ver, browser_name, ghz)

        ram = virtual_memory()

        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect(("8.8.8.8", 80))
        ip = s.getsockname()[0]
        s.close()

        client = self.pt_suite.addNode(ptVM("client", version=platform.platform(), ip=ip, cpus=cpu_count(),
                                       ram_mb=int(ram.total / (1024 * 1024)) if ram else 0))
        client.addNode(ptComponent(self.browser.browser_get_name(), self.browser.browser_get_version()))

    def _pt_upload(self):
        if not self.opts.pt_project:
            return

        self.pt_suite.upload()

    def _pt_update_results(self, page):
        if self.opts.pt_project is None or page is None:
            return

        basename = page.get_full_name()
        if basename is None:
            return

        for s, m, f, g in (("Rendering", "msec", "dur", "Page rendering"),
                           ("Footprint", "KB", "ram_usage_kb", "Browser memory footprint")):
            name = "%s [%s, %s]" % (basename, s, m)
            group = "%s (%s)" % (g, m)
            test = self.pt_suite.getTest(name, group=group)
            if test is None:
                test = ptTest(name, less_better=True, group=group, metrics=m, description=page.url)
                self.pt_suite.addTest(test)

            test.add_score(page.__dict__[f])
            test.duration_sec += int(math.ceil(page.dur))

        self._pt_upload()

    def _run(self):

        self.browser.print_browser_info()

        urls = copy.copy(self.urls)

        CP = None

        if self.user:
            if self.opts.uncached:
                logging.error("Can't combine --user and --uncached mode")
                return None
            if len(self.urls) == 0:
                logging.error("Login URL is not specified")
                return None

            self._login(urls[0][1], wait_completion=self.opts.login_wait and self.opts.add_login_landing)

            CP = self._detect_cp_type()

            if self.opts.add_login_landing:
                curr_url = self.browser.browser_get_current_url()
                if curr_url not in [u[1] for u in urls]:
                    urls = [('Login landing page', curr_url)] + urls

        if self.opts.menu_walk:
            if not self.user:
                name, url = urls[0]
                page = self.browser.navigate_to(url, name=name)

            if CP is None:
                CP = self._detect_cp_type()

            if not CP:
                logging.error("Can't recognize Control Panel, aborting")
                sys.exit(-1)
            items = CP.cp_do_scan_menu()
            if items:
                urls = items

        if CP is None:
            CP = CPEngineBase(self.browser, html_report=self._html_report)

        CP.cp_handle_opts(self.opts)
        self._pt_suite_init(CP)

        if self.opts.randomize_urls:
            import random
            for i in range(len(urls)):
                j = random.randint(0, len(urls) - 1)
                urls[i], urls[j] = urls[j], urls[i]

        print("")
        PageStats.print_title("URLs to navigate")
        print("  " + "\n  ".join(["%s%s" % (("%-25s - " % u[0]) if u[0] else "", u[1]) for u in urls]))

        pages = {}

        cached = not self.opts.uncached

        print_title = True
        for name, url in urls:
            try:
                page = CP.cp_do_navigate(url, cached=cached, name=name)
                self._pt_update_results(page)
                if not page:
                    raise CPCrawlerException("Page navigation (%s) failed, aborting" % url)

                pages[url] = page
            except BrowserExc as e:
                logging.error("Browser exception @ %s: %s\n%s" % (name, url, e))
                continue

            if self.opts.html_report:
                img_path = self._html_report.add_page(url, page=page)
                self.browser.browser_get_screenshot_as_file(img_path)
                self._html_report.gen_thumbnails(url)

            if self.opts.requests:
                description = ["SCREEN: %s" % page.get_full_name(), "URL: %s" % page.url]
                if print_title:
                    page.print_page_req_groups_stats(title=print_title, description=description)
                else:
                    page.print_page_req_groups_stats(title=print_title, description=description)
                print_title = False

        if self.opts.requests:
            print_title = True
            for page in pages.values():
                description = ["SCREEN: %s" % page.get_full_name(), "URL: %s" % page.url]
                page.print_page_requests_stats(title=print_title, description=description)
                print_title = False

        need_to_exit = False

        browser_page_stats = []
        simulators_page_stats = []

        if self.opts.python_browsers:
            simulators = []
            for n in range(0, self.opts.python_browsers):
                log_path = os.path.join(self.logdir, "%s.%d.log" % (BrowserPython.engine, n))
                simulators.append(BrowserPython(log_path=log_path))

        for name, url in urls:
            if url not in pages:
                continue

            page = pages[url]

            if self.opts.python_browsers:
                map(lambda x: x.loop_start([page], sleep_sec=self.opts.delay), simulators)

            description = ["BROWSER:      %s" % self.browser.browser_get_name(),
                           "SIMULATORS:   %d browser(s) in background" % (self.opts.python_browsers),
                           ] if self.opts.python_browsers else []

            page_full_name = page.get_full_name()
            page_url = page.url

            description.append("%11s %s" % ("SCREEN (CACHED):" if cached else "SCREEN (UNCACHED):", page_full_name))
            if page_url != page_full_name:
                description.append("%11s %s" % ("URL:", page_url))

            self.browser.page_stats[page.get_key()].print_page_timeline_header(title=not len(browser_page_stats),
                                                                               description=description)

            try:
                for n in range(0, self.opts.loops):
                    if n == 0 and not self.opts.python_browsers:
                        # use already fetched page
                        self.browser.page_stats[page.get_key()].print_page_timeline(pages[url], title=str(n + 1))
                        continue

                    time.sleep(self.opts.delay)
                    try:
                        if self.opts.reset_dom:
                            CP.browser.navigation_reset()
                        page = CP.cp_do_navigate(url, cached=cached, name=name)
                        self._pt_update_results(page)
                        if not page:
                            raise CPCrawlerException("Page navigation (%s) failed, aborting" % url)

                        self.browser.page_stats[page.get_key()].print_page_timeline(page, title=str(n + 1))
                    except BrowserExc as e:
                        logging.error(e)
                        # break
            except KeyboardInterrupt:
                need_to_exit = True

            br_ps = self.browser.page_stats[page.get_key()]
            br_ps.id = self.browser.browser_get_name()
            avg = br_ps.get_avg()
            self.browser.page_stats[page.get_key()].print_page_timeline(avg, title="Average", hr=True)
            browser_page_stats.append(br_ps)

            if self.opts.python_browsers:
                map(lambda x: x.loop_stop(), simulators)
                map(lambda x: x.loop_wait(), simulators)

                py_ps = PageStats("%d python simulator(s)" % (self.opts.python_browsers))
                for s in simulators:
                    py_ps.iterations += s.page_stats[page.get_key()].iterations

                simulators_page_stats.append(py_ps)

            if need_to_exit:
                break

        for ps in browser_page_stats:
            self.page_stats_summary.add_page_stats(ps)
        for ps in simulators_page_stats:
            self.page_stats_summary.add_page_stats(ps)

    def run(self):

        if self.opts.log2file or self.opts.log_file:
            logging.basicConfig(filename=self.crawler_logfile, level=logging.DEBUG)
            print("Redirecting %sverbose logs to %s" %
                  ("browser.%d " % self.browser_id if self.browser_id is not None else "", self.crawler_logfile))
        else:
            level = logging.DEBUG if self.opts.verbose > 1 else logging.INFO if self.opts.verbose else logging.WARNING
            logging.basicConfig(level=level)

        logger = logging.getLogger()
        for handler in logger.root.handlers:
            handler.setFormatter(CPLogFormatter())

        self.init()

        if self.opts.session:
            self.browser.domain_set_session(self.urls[0], self.opts.session)

        try:
            self._run()
            print("")
        except KeyboardInterrupt:
            pass
        except BrowserExc as e:
            print(e)
        except RuntimeError:
            self.fini()
            raise
        finally:
            self.browser.browser_stop()
            self.fini()

        if not self.browser_id:
            if self.page_stats_summary:
                self.page_stats_summary.print_summary(title='Final summary')

            if self.opts.wait:
                print("Press Ctrl+C to exit...")
                try:
                    time.sleep(10000)
                except KeyboardInterrupt:
                    pass
        else:
            print("browser.%d done" % self.browser_id)


class CPCrawler:
    def __init__(self, workdir=None, logfile=None, pt_job_title=None, pt_server_url=None):

        self.workdir = workdir if workdir else os.path.join(gettempdir(), "%s.%d" % (basename, os.getpid()))
        self.logfile = logfile
        self.opts = None
        self.urls = []

        self.pt_suite = ptSuite("UI crawler job" if pt_job_title is None else pt_job_title,
                                pt_server_url=pt_server_url)

    def _gen_users(self, users_ar):
        if not users_ar:
            return None

        users = []
        for u in users_ar:
            for _u in u.split(","):
                if "{" in _u:
                    m = re.search(r"(?P<pfx>.*?){(?P<from>\d+)-(?P<to>\d+)}(?P<sfx>.*)", _u)
                    if not m:
                        raise CPCrawlerException("can't parse users range from '%s',"
                                                 "valid pattern is 'prefix{from-to}suffix'" % _u)
                    for n in range(int(m.group('from')), int(m.group('to')) + 1):
                        users.append(m.group('pfx') + str(n) + m.group('sfx'))
                else:
                    users.append(_u)
        return users

    def add_options(self, op, passwd="password",
                    nav_timeout=OPT_DEFAULT_NAV_TIMEOUT,
                    ajax_threshold=OPT_DEFAULT_AJAX_THRESHOLD,
                    delay_between_click_sec=OPT_DELAY_BETWEEN_CLICK_SEC):

        og = OptionGroup(op, "Control Panel crawler options")
        og.add_option("-v", "--verbose", action="count", default=0,
                      help="run in verbose mode, use -vvv for max verbosity")
        og.add_option("", "--log-file", type="string", help="log into the file (default %s)" % self.logfile)
        og.add_option("", "--log2file", action="store_true", help="enable --verbose mode and log to the --log-file")
        og.add_option("-V", "--view", action="store_true", help="Show browser screen")
        og.add_option("--remote-connstring", default=None, type="string",
                      help="Connect to remote selenium. ex. http://{SELENIUM_IP}:{SELENIUM_PORT}/wd/hub")
        og.add_option("-g", "--html-report", type="string",
                      help="generate HTML report with screenshots and other information")
        og.add_option("-t", "--telemetry", type="string",
                      help="log pages to given file (append only, concurrent-process-safe)")
        og.add_option("-w", "--wait", action="store_true", help="don\'t close the browser and wait till test is killed")
        og.add_option("-l", "--loops", type="int", default=7, help="number of iterations, default %default")

        og.add_option("-b", "--browser", choices=[b.engine for b in BROWSERS], default=BROWSERS[0].engine,
                      help="browser to use: %s (default is '%%default')" %
                      ",".join(['\'%s\'' % b.engine for b in BROWSERS]))
        og.add_option("-r", "--requests", action="store_true",
                      help="print information about individual network requests")
        og.add_option("-o", "--perf-atomic-format", action="store_true", help="perf-atomic output format")
        op.add_option_group(og)

        og = OptionGroup(op, "Authentication options")
        og.add_option("-U", "--user", action="append", type="string", default=None,
                      help="try to login with given user name before the test (comma-separated list accepted)")
        og.add_option("-P", "--password", type="string", default=passwd, help="password, default: %default")
        op.add_option_group(og)

        og = OptionGroup(op, "Navigation options")
        og.add_option("-m", "--menu-walk", action="store_true",
                      help="search for menu items and click on every menu item")
        og.add_option("-L", "--skip-login-landing-url", action="store_false", dest="add_login_landing", default=True,
                      help="skip login landing page URL")
        og.add_option("", "--dont-wait-login-landing", action="store_false", dest="login_wait", default=True,
                      help="do not wait for full login landing page load")
        og.add_option("-R", "--randomize-urls", action="store_true", help="randomize menu items sequence")
        og.add_option("-f", "--urls-file", type='string', help="get URLs from file (instead of command line)")
        og.add_option("-i", "--dir-index", action="store_true",
                      help="treat given page as apache directory listing index page and parse URLs from there")
        og.add_option("-u", "--uncached", action="store_true", help="invalidate browser cache before each request")
        og.add_option("-s", "--session", type="string", help="session ID")
        og.add_option("-e", "--reset-dom", action="store_true", help="reset DOM model after every nav click")

        op.add_option_group(og)

        og = OptionGroup(op, "Page navigation timings")
        og.add_option("-d", "--delay", type="float", default=delay_between_click_sec,
                      help="delay between GET requests, default %default sec")
        og.add_option("-T", "--nav-timeout", type="float", dest='nav_timeout', default=nav_timeout,
                      help="navigation timeout, sec (default %default)")
        og.add_option("-A", "--ajax-threshold", type="float", dest='ajax_threshold', default=ajax_threshold,
                      help="page considered as complete if no ajax requests issued after the "
                           "AJAX_THRESHOLD, sec (default %default)")
        op.add_option_group(og)

        og = OptionGroup(op, "Mass load mode")
        og.add_option("-X", "--real-browsers", type="int", default=0,
                      help="run REAL_BROWSERS instances of the real browsers, users (if any) "
                           "will be spread equally by the instances")
        og.add_option("-x", "--python-browsers", type="int", default=0,
                      help="run PYTHON_BROWSERS satellite/clone python browsers per every real "
                           "browser to increase load on server")
        og.add_option("-D", "--instances-delay", type="float", default=0.3,
                      help="delay between instances start (sec) (default %default)")
        og.add_option("-W", "--work-dir", type="string",
                      help="work directory with the tool logs and final HTML report, default %default")
        op.add_option_group(og)

        self.pt_suite.addOptions(op)

    def init_opts(self, opts, urls, init_logging=True):

        if opts.urls_file:
            if not urls:
                urls = []
            try:
                f = open(opts.urls_file)
                for line in f.readlines():
                    urls.append(line.strip())
                f.close()
            except Exception as e:
                raise CPCrawlerException("ERROR: %s" % e)

        if not urls:
            raise CPCrawlerException("URL is not specified")

        if opts.work_dir:
            self.workdir = opts.work_dir
            self.logfile = basename + ".log"

        if opts.log_file:
            self.logfile = opts.log_file
        self.opts = opts
        if opts.dir_index:
            self.urls = [(None, u) for u in gen_urls_from_index_file(urls)]
        else:
            self.urls = [(None, u) for u in urls]

        shutil.rmtree(self.workdir, ignore_errors=True)
        os.makedirs(self.workdir, mode=0o777)

        if init_logging:
            self.init_logging(opts)

        self.pt_suite.handleOptions(opts)

    def init_logging(self, opts):
        if opts.log2file or opts.log_file:
            logging.basicConfig(filename=self.logfile, level=logging.DEBUG)
            print("Redirecting verbose logs to %s" % self.logfile)
        else:
            level = logging.DEBUG if opts.verbose > 1 else logging.INFO if opts.verbose else logging.WARNING
            logging.basicConfig(level=level)
            if opts.verbose < 3:
                selenium_logger.setLevel(logging.WARNING)

        logger = logging.getLogger()
        for handler in logger.root.handlers:
            handler.setFormatter(CPLogFormatter())

    def crawl(self, cp_engines=None):

        def _browser_launch(queue, cp_engines, opts, urls, users, browser_id, logfile, workdir, pt_suite):
            cpbr = CPBrowserRunner(cp_engines, opts, urls, users, browser_id, logfile, workdir, pt_suite)
            if browser_id:
                try:
                    cpbr.run()
                except RuntimeError:
                    logging.error(traceback.format_exc())
            else:
                cpbr.run()
            if queue:
                queue.put(cpbr)
            return cpbr

        if not cp_engines:
            cp_engines = [CPEngineBase]

        if self.opts is None:
            raise CPCrawlerException("the init_opts() method must be called before run()")

        users = self._gen_users(self.opts.user)

        if users and len(users) > 1 and len(users) > self.opts.real_browsers:
            raise CPCrawlerException("ERROR: number of users (%d) can't be higher than number of browsers (%d)"
                                     % (len(users), self.opts.real_browsers))

        if self.opts.real_browsers > 0:
            real_browsers = []
            cpbr_objs = {}

            for i in range(1, self.opts.real_browsers + 1):
                cpbr_objs[i] = Queue()
                b = Process(target=_browser_launch, args=(cpbr_objs[i], cp_engines, self.opts, self.urls, users, i,
                                                          self.logfile, self.workdir, self.pt_suite))
                b.start()
                real_browsers.append(b)
                time.sleep(self.opts.instances_delay)

            delay = 3
            telemetry_fname = os.path.join(self.workdir, "telemetry.log")
            report_fname = os.path.join(self.workdir, "report.html")

            print("")
            print("Notes:")
            print("   * Tests executed by %d %s browsers%s%s" %
                  (self.opts.real_browsers, self.opts.browser,
                   " and %d python browsers" % (self.opts.real_browsers * self.opts.python_browsers)
                   if self.opts.python_browsers else "",
                   " on %d users:\n       %s" % (len(users), "\n       ".join(users)) if users else ""))
            print("   * All data is stored in %s" % self.workdir)
            print("   * The %s file is updated every %d sec" % (report_fname, delay))
            print("")

            while True:
                time.sleep(delay)
                all_dead = True
                for b in real_browsers:
                    if not b.is_alive():
                        continue
                    all_dead = False

                if all_dead:
                    break

            page_stats_summary = PageStatsSummary()

            for i in range(1, self.opts.real_browsers + 1):
                cpbr = cpbr_objs[i].get()
                if os.path.exists(cpbr.stderr_fname):
                    f = open(cpbr.stderr_fname, 'r')
                    data = f.read()
                    f.close()
                    if data:
                        print("browser.%d stderr:\n%s" % (i, data))

                for ps in cpbr.page_stats_summary.page_stats:
                    page_stats_summary.add_page_stats(ps)

            report_fname = os.path.join(self.workdir, "report.html")

        else:
            cpbr = _browser_launch(None, cp_engines, self.opts, self.urls, users, 0, self.logfile,
                                   self.workdir, self.pt_suite)
            page_stats_summary = cpbr.page_stats_summary

        return page_stats_summary.page_stats


##############################################################################
# Autotests
##############################################################################


if __name__ == "__main__":
    usage = "usage: %prog [options] URL [URL2 [URL3 ...]]"

    op = OptionParser(usage=usage)
    og = OptionGroup(op, "My specific options")
    op.add_option("-c", "--config", type="string", help="Some config file")
    op.add_option_group(og)

    cpc = CPCrawler()
    cpc.add_options(op)

    opts, urls = op.parse_args()

    # override options to simplify the test
    opts.loops = 1
    urls = ["https://example.com/"]

    cpc.init_opts(opts, urls)

    cpc.crawl()
